Compare commits
21 Commits
fcf531a9b2
...
main
| Author | SHA1 | Date | |
|---|---|---|---|
| 6309dc1181 | |||
| 264c004662 | |||
| 2fe903ecea | |||
| aac9ace144 | |||
| 6da0972740 | |||
| 40d8a29e33 | |||
| fd392f7fbb | |||
| 10a98539d0 | |||
| cc3bc2188c | |||
| 06a798cab9 | |||
| 9a1af0adee | |||
| d2c55c47b2 | |||
| 14925154a3 | |||
| a24621fa6a | |||
| 68b55fa1e6 | |||
| 8f11d6e5cd | |||
| e04a8ffb18 | |||
| 6485c87c5b | |||
| a77239c0c8 | |||
| e5734b41fa | |||
| 42e13f33dd |
@@ -83,6 +83,24 @@ unsafe extern "C" {
|
||||
scale: f32,
|
||||
stream: *mut c_void,
|
||||
);
|
||||
fn launch_paged_decode_attention_tree_bf16(
|
||||
q: *const c_void,
|
||||
k_cache: *const c_void,
|
||||
v_cache: *const c_void,
|
||||
o: *mut c_void,
|
||||
block_tables: *const i32,
|
||||
context_lens: *const i32,
|
||||
tree_mask: *const i32,
|
||||
batch: i32,
|
||||
num_q_heads: i32,
|
||||
num_kv_heads: i32,
|
||||
head_dim: i32,
|
||||
max_blocks_per_seq: i32,
|
||||
tree_start: i32,
|
||||
tree_len: i32,
|
||||
scale: f32,
|
||||
stream: *mut c_void,
|
||||
);
|
||||
fn launch_paged_decode_attention_sinks_bf16(
|
||||
q: *const c_void,
|
||||
k_cache: *const c_void,
|
||||
@@ -127,6 +145,17 @@ unsafe extern "C" {
|
||||
max_blocks_per_seq: i32,
|
||||
stream: *mut c_void,
|
||||
);
|
||||
fn launch_copy_kv_position(
|
||||
k_pool: *mut c_void,
|
||||
v_pool: *mut c_void,
|
||||
block_ids: *const i32,
|
||||
src_pos: i32,
|
||||
dst_pos: i32,
|
||||
num_kv_heads: i32,
|
||||
head_dim: i32,
|
||||
block_size: i32,
|
||||
stream: *mut c_void,
|
||||
);
|
||||
}
|
||||
|
||||
/// Scatter `[num_kv_heads, num_tokens, head_dim]` BF16 K/V into a paged
|
||||
@@ -213,6 +242,36 @@ pub unsafe fn reshape_and_cache_batched_bf16(
|
||||
}
|
||||
}
|
||||
|
||||
/// Copy one token's K/V from `src_pos` to `dst_pos` within the same sequence's
|
||||
/// paged cache (one layer). Used by tree speculative decoding to remap
|
||||
/// accepted sibling K/V to canonical sequential positions after acceptance.
|
||||
///
|
||||
/// # Safety
|
||||
/// Pool and block_ids pointers must be valid GPU pointers for the given layer.
|
||||
pub unsafe fn copy_kv_position(
|
||||
k_pool_ptr: *mut c_void,
|
||||
v_pool_ptr: *mut c_void,
|
||||
block_ids_gpu: *const i32,
|
||||
src_pos: usize,
|
||||
dst_pos: usize,
|
||||
num_kv_heads: usize,
|
||||
head_dim: usize,
|
||||
block_size: usize,
|
||||
stream: *mut c_void,
|
||||
) {
|
||||
launch_copy_kv_position(
|
||||
k_pool_ptr,
|
||||
v_pool_ptr,
|
||||
block_ids_gpu,
|
||||
src_pos as i32,
|
||||
dst_pos as i32,
|
||||
num_kv_heads as i32,
|
||||
head_dim as i32,
|
||||
block_size as i32,
|
||||
stream,
|
||||
);
|
||||
}
|
||||
|
||||
fn apply_causal_mask(scores: &Tensor, offset: usize) {
|
||||
let ndim = scores.ndim();
|
||||
let rows = scores.shape()[ndim - 2];
|
||||
@@ -515,6 +574,62 @@ pub fn paged_decode_attention(
|
||||
output
|
||||
}
|
||||
|
||||
/// Tree-aware paged decode attention. Adds a per-query attention mask over
|
||||
/// the newly-written K/V region `[tree_start, tree_start+tree_len)`. Query i
|
||||
/// attends to position tree_start+j iff tree_mask[i, j] != 0. Positions <
|
||||
/// tree_start are always attended.
|
||||
///
|
||||
/// Used by speculative decoding with tree drafting to let sibling candidates
|
||||
/// share position slots without seeing each other's K/V.
|
||||
#[allow(clippy::too_many_arguments)]
|
||||
pub fn paged_decode_attention_tree(
|
||||
q: &Tensor,
|
||||
k_cache_ptr: *const c_void,
|
||||
v_cache_ptr: *const c_void,
|
||||
block_tables_ptr: *const i32,
|
||||
context_lens_ptr: *const i32,
|
||||
tree_mask_ptr: *const i32,
|
||||
batch: usize,
|
||||
num_q_heads: usize,
|
||||
num_kv_heads: usize,
|
||||
head_dim: usize,
|
||||
max_blocks_per_seq: usize,
|
||||
tree_start: usize,
|
||||
tree_len: usize,
|
||||
) -> Tensor {
|
||||
assert_eq!(q.ndim(), 4);
|
||||
assert_eq!(q.shape()[2], 1);
|
||||
assert_eq!(q.dtype(), DType::BF16);
|
||||
assert!(num_q_heads % num_kv_heads == 0);
|
||||
assert!(head_dim <= 128);
|
||||
|
||||
let scale = 1.0 / (head_dim as f32).sqrt();
|
||||
let output = Tensor::empty(&[batch, num_q_heads, 1, head_dim], DType::BF16, q.device());
|
||||
|
||||
unsafe {
|
||||
launch_paged_decode_attention_tree_bf16(
|
||||
q.data_ptr() as *const c_void,
|
||||
k_cache_ptr,
|
||||
v_cache_ptr,
|
||||
output.data_ptr() as *mut c_void,
|
||||
block_tables_ptr,
|
||||
context_lens_ptr,
|
||||
tree_mask_ptr,
|
||||
batch as i32,
|
||||
num_q_heads as i32,
|
||||
num_kv_heads as i32,
|
||||
head_dim as i32,
|
||||
max_blocks_per_seq as i32,
|
||||
tree_start as i32,
|
||||
tree_len as i32,
|
||||
scale,
|
||||
xserv_cuda::current_stream_raw(),
|
||||
);
|
||||
}
|
||||
|
||||
output
|
||||
}
|
||||
|
||||
/// Paged decode attention with attention sinks and optional sliding window.
|
||||
///
|
||||
/// sinks_ptr: pointer to [num_q_heads] BF16 on GPU (or null for no sinks)
|
||||
|
||||
@@ -18,6 +18,17 @@ unsafe extern "C" {
|
||||
n: i32,
|
||||
stream: *mut c_void,
|
||||
);
|
||||
|
||||
fn launch_gemv_bf16_batched(
|
||||
x: *const c_void,
|
||||
w: *const c_void,
|
||||
y_bf16: *mut c_void,
|
||||
y_fp32_buf: *mut c_void,
|
||||
m: i32,
|
||||
k: i32,
|
||||
n: i32,
|
||||
stream: *mut c_void,
|
||||
);
|
||||
}
|
||||
|
||||
#[derive(Debug, Clone, Copy)]
|
||||
@@ -31,6 +42,55 @@ pub fn gemv_scratch_elems(k: usize, n: usize) -> usize {
|
||||
n * k.div_ceil(GEMV_TILE_K)
|
||||
}
|
||||
|
||||
/// Batched GEMV: [M, K] × [K, N] → [M, N], all BF16.
|
||||
/// Bit-exact with calling matmul on each row individually (same K-block partial
|
||||
/// + fixed-order reduction path), but in a single kernel launch per phase.
|
||||
pub fn matmul_batched_gemv(a: &Tensor, b: &Tensor) -> Tensor {
|
||||
assert_eq!(a.ndim(), 2);
|
||||
assert_eq!(b.ndim(), 2);
|
||||
assert!(a.is_contiguous());
|
||||
assert!(b.is_contiguous());
|
||||
assert_eq!(a.dtype(), DType::BF16);
|
||||
assert_eq!(b.dtype(), DType::BF16);
|
||||
let m = a.shape()[0];
|
||||
let k = a.shape()[1];
|
||||
let n = b.shape()[1];
|
||||
assert_eq!(b.shape()[0], k);
|
||||
|
||||
let out = Tensor::empty(&[m, n], DType::BF16, a.device());
|
||||
let scratch_elems = m * gemv_scratch_elems(k, n);
|
||||
let mut fp32_buf = xserv_cuda::allocator::cached_alloc(scratch_elems * 4).unwrap();
|
||||
|
||||
let null_stream = xserv_cuda::current_stream_raw();
|
||||
if m == 1 {
|
||||
unsafe {
|
||||
launch_gemv_bf16(
|
||||
a.data_ptr() as *const c_void,
|
||||
b.data_ptr() as *const c_void,
|
||||
out.data_ptr() as *mut c_void,
|
||||
fp32_buf.as_mut_ptr() as *mut c_void,
|
||||
k as i32,
|
||||
n as i32,
|
||||
null_stream,
|
||||
);
|
||||
}
|
||||
} else {
|
||||
unsafe {
|
||||
launch_gemv_bf16_batched(
|
||||
a.data_ptr() as *const c_void,
|
||||
b.data_ptr() as *const c_void,
|
||||
out.data_ptr() as *mut c_void,
|
||||
fp32_buf.as_mut_ptr() as *mut c_void,
|
||||
m as i32,
|
||||
k as i32,
|
||||
n as i32,
|
||||
null_stream,
|
||||
);
|
||||
}
|
||||
}
|
||||
out
|
||||
}
|
||||
|
||||
// --- FFI: custom CUDA kernels ---
|
||||
unsafe extern "C" {
|
||||
fn launch_gemm_naive_f32(
|
||||
|
||||
@@ -15,11 +15,12 @@ pub mod transpose;
|
||||
pub use activation::{add, bias_add_2d, gelu, gpt_oss_glu, mul, scale, silu, silu_mul};
|
||||
pub use argmax::{argmax_bf16_single, argmax_bf16_to_host};
|
||||
pub use attention::{
|
||||
attention, decode_attention, flash_attention, flash_attention_sinks, paged_decode_attention,
|
||||
paged_decode_attention_sinks, reshape_and_cache_batched_bf16, reshape_and_cache_bf16,
|
||||
attention, copy_kv_position, decode_attention, flash_attention, flash_attention_sinks,
|
||||
paged_decode_attention, paged_decode_attention_sinks, paged_decode_attention_tree,
|
||||
reshape_and_cache_batched_bf16, reshape_and_cache_bf16,
|
||||
};
|
||||
pub use embedding::{embedding, embedding_device_ids};
|
||||
pub use gemm::{GemmBackend, batched_matmul, matmul};
|
||||
pub use gemm::{GemmBackend, batched_matmul, matmul, matmul_batched_gemv};
|
||||
pub use layernorm::layernorm;
|
||||
pub use rmsnorm::{add_rmsnorm, rmsnorm};
|
||||
pub use rope::{RopeCache, rope_inplace, rope_inplace_device_pos};
|
||||
|
||||
1126
crates/xserv-model/src/bin/bench-eagle3.rs
Normal file
1126
crates/xserv-model/src/bin/bench-eagle3.rs
Normal file
File diff suppressed because it is too large
Load Diff
@@ -10,6 +10,7 @@ use half::bf16;
|
||||
use std::path::{Path, PathBuf};
|
||||
use std::time::Instant;
|
||||
|
||||
use xserv_model::qwen3_graph::GraphedQwen3Decoder;
|
||||
use xserv_model::{BLOCK_SIZE, ModelConfig, PagedKVCache, Qwen3, loader};
|
||||
use xserv_tensor::{DType, Device, Tensor};
|
||||
use xserv_tokenizer::Tokenizer;
|
||||
@@ -222,12 +223,14 @@ fn main() {
|
||||
let mut target_verify_cache =
|
||||
new_cache_with_rows(&target_config, max_seq_len, device, gamma);
|
||||
let mut draft_cache = new_cache(&draft_config, max_seq_len, device);
|
||||
let mut draft_decoder = GraphedQwen3Decoder::new();
|
||||
let _ = run_speculative(
|
||||
&target,
|
||||
&draft,
|
||||
&mut target_cache,
|
||||
&mut target_verify_cache,
|
||||
&mut draft_cache,
|
||||
&mut draft_decoder,
|
||||
&tokenizer,
|
||||
&warm_ids,
|
||||
warm_tokens,
|
||||
@@ -248,6 +251,21 @@ fn main() {
|
||||
);
|
||||
|
||||
let mut totals = Totals::default();
|
||||
|
||||
// Persistent per-benchmark caches so the draft CUDA graph (Phase 24) can be
|
||||
// captured once and replayed across every prompt. Freeing and re-registering
|
||||
// slot 0 between prompts keeps block_table_gpu / context_lens_gpu addresses
|
||||
// stable, which is exactly what the graph captured.
|
||||
let mut target_cache = new_cache_with_rows(
|
||||
&target_config,
|
||||
max_seq_len,
|
||||
device,
|
||||
if use_verify_logits { gamma } else { 1 },
|
||||
);
|
||||
let mut target_verify_cache = new_cache_with_rows(&target_config, max_seq_len, device, gamma);
|
||||
let mut draft_cache = new_cache(&draft_config, max_seq_len, device);
|
||||
let mut draft_decoder = GraphedQwen3Decoder::new();
|
||||
|
||||
for (i, prompt) in PROMPTS.iter().take(prompt_count).enumerate() {
|
||||
let ids = tokenizer.encode(prompt);
|
||||
validate_length_budget(&ids, gen_tokens, max_seq_len, prompt);
|
||||
@@ -255,21 +273,13 @@ fn main() {
|
||||
let baseline = run_baseline(&target, &mut baseline_cache, &tokenizer, &ids, gen_tokens);
|
||||
drop(baseline_cache);
|
||||
|
||||
let mut target_cache = new_cache_with_rows(
|
||||
&target_config,
|
||||
max_seq_len,
|
||||
device,
|
||||
if use_verify_logits { gamma } else { 1 },
|
||||
);
|
||||
let mut target_verify_cache =
|
||||
new_cache_with_rows(&target_config, max_seq_len, device, gamma);
|
||||
let mut draft_cache = new_cache(&draft_config, max_seq_len, device);
|
||||
let spec = run_speculative(
|
||||
&target,
|
||||
&draft,
|
||||
&mut target_cache,
|
||||
&mut target_verify_cache,
|
||||
&mut draft_cache,
|
||||
&mut draft_decoder,
|
||||
&tokenizer,
|
||||
&ids,
|
||||
gen_tokens,
|
||||
@@ -438,6 +448,7 @@ fn run_speculative(
|
||||
target_cache: &mut PagedKVCache,
|
||||
target_verify_cache: &mut PagedKVCache,
|
||||
draft_cache: &mut PagedKVCache,
|
||||
draft_decoder: &mut GraphedQwen3Decoder,
|
||||
tokenizer: &Tokenizer,
|
||||
prompt_ids: &[u32],
|
||||
gen_tokens: usize,
|
||||
@@ -504,7 +515,7 @@ fn run_speculative(
|
||||
break;
|
||||
}
|
||||
let pos = draft_cache.seq_len(slot);
|
||||
let logits = draft.forward_decode_paged(&[token], &[pos], &[slot], draft_cache);
|
||||
let logits = draft_decoder.decode(draft, &[token], &[pos], &[slot], draft_cache);
|
||||
draft_next = last_argmax(&logits);
|
||||
}
|
||||
proposed_total += draft_tokens.len();
|
||||
@@ -572,6 +583,7 @@ fn run_speculative(
|
||||
.unwrap();
|
||||
replay_draft_tokens(
|
||||
draft,
|
||||
draft_decoder,
|
||||
draft_cache,
|
||||
slot,
|
||||
&draft_tokens[..accepted],
|
||||
@@ -588,7 +600,7 @@ fn run_speculative(
|
||||
commit_steps += 1;
|
||||
|
||||
let pos = draft_cache.seq_len(slot);
|
||||
let logits = draft.forward_decode_paged(&[correction], &[pos], &[slot], draft_cache);
|
||||
let logits = draft_decoder.decode(draft, &[correction], &[pos], &[slot], draft_cache);
|
||||
draft_next = last_argmax(&logits);
|
||||
correction_steps += 1;
|
||||
continue;
|
||||
@@ -690,6 +702,7 @@ fn run_speculative(
|
||||
.unwrap();
|
||||
replay_draft_tokens(
|
||||
draft,
|
||||
draft_decoder,
|
||||
draft_cache,
|
||||
slot,
|
||||
&draft_tokens[..accepted],
|
||||
@@ -709,7 +722,7 @@ fn run_speculative(
|
||||
mirror_steps += 1;
|
||||
|
||||
let pos = draft_cache.seq_len(slot);
|
||||
let logits = draft.forward_decode_paged(&[correction], &[pos], &[slot], draft_cache);
|
||||
let logits = draft_decoder.decode(draft, &[correction], &[pos], &[slot], draft_cache);
|
||||
draft_next = last_argmax(&logits);
|
||||
correction_steps += 1;
|
||||
}
|
||||
@@ -745,6 +758,7 @@ fn advance_target_cache(target: &Qwen3, cache: &mut PagedKVCache, slot: usize, t
|
||||
|
||||
fn replay_draft_tokens(
|
||||
draft: &Qwen3,
|
||||
draft_decoder: &mut GraphedQwen3Decoder,
|
||||
cache: &mut PagedKVCache,
|
||||
slot: usize,
|
||||
tokens: &[u32],
|
||||
@@ -752,7 +766,7 @@ fn replay_draft_tokens(
|
||||
) {
|
||||
for &token in tokens {
|
||||
let pos = cache.seq_len(slot);
|
||||
let logits = draft.forward_decode_paged(&[token], &[pos], &[slot], cache);
|
||||
let logits = draft_decoder.decode(draft, &[token], &[pos], &[slot], cache);
|
||||
*next = last_argmax(&logits);
|
||||
}
|
||||
}
|
||||
|
||||
134
crates/xserv-model/src/bin/bench-verify-cost.rs
Normal file
134
crates/xserv-model/src/bin/bench-verify-cost.rs
Normal file
@@ -0,0 +1,134 @@
|
||||
//! Micro-benchmark: measure the cost of forward_verify_paged_decode_attention
|
||||
//! at different batch sizes (γ+1 values), to understand where speedup comes
|
||||
//! from (or doesn't).
|
||||
|
||||
use std::path::PathBuf;
|
||||
use std::time::Instant;
|
||||
|
||||
use xserv_model::{BLOCK_SIZE, ModelConfig, PagedKVCache, Qwen3, loader};
|
||||
use xserv_tensor::{DType, Device};
|
||||
use xserv_tokenizer::Tokenizer;
|
||||
|
||||
fn main() {
|
||||
let args: Vec<String> = std::env::args().collect();
|
||||
if args.len() < 2 {
|
||||
eprintln!(
|
||||
"Usage: bench-verify-cost <target-dir> [--prompt-len N] [--iters N] [--device N]"
|
||||
);
|
||||
std::process::exit(1);
|
||||
}
|
||||
let target_dir = PathBuf::from(&args[1]);
|
||||
let prompt_len = arg_usize(&args, "--prompt-len", 100);
|
||||
let iters = arg_usize(&args, "--iters", 30);
|
||||
let device = arg_usize(&args, "--device", 0) as u32;
|
||||
|
||||
xserv_cuda::device::set_device(device).unwrap();
|
||||
|
||||
let cfg = ModelConfig::from_file(&target_dir.join("config.json"));
|
||||
eprintln!("Loading target...");
|
||||
let weights = loader::load_model_dir(&target_dir, Device::Cuda(device));
|
||||
let target = Qwen3::from_weights(cfg.clone(), weights);
|
||||
xserv_cuda::allocator::cached_trim();
|
||||
|
||||
let tok = Tokenizer::from_file(&target_dir.join("tokenizer.json"));
|
||||
let ids = tok.encode(&"the ".repeat(prompt_len))[..prompt_len].to_vec();
|
||||
|
||||
let max_seq_len = 2048;
|
||||
let num_blocks = (max_seq_len + BLOCK_SIZE - 1) / BLOCK_SIZE + 4;
|
||||
let mut cache = PagedKVCache::new(&cfg, num_blocks, 0, 16, num_blocks, DType::BF16, device);
|
||||
cache.register_sequence(0).unwrap();
|
||||
|
||||
// Prefill
|
||||
let _ = target.forward_prefill_paged(&ids, 0, &mut cache);
|
||||
sync();
|
||||
|
||||
// Warmup one of each
|
||||
for &n in &[1, 2, 3, 5, 9] {
|
||||
let toks: Vec<u32> = (0..n).map(|_| ids[0]).collect();
|
||||
let _ = target.forward_decode_paged(
|
||||
&toks,
|
||||
&(0..n).map(|i| ids.len() + i).collect::<Vec<_>>(),
|
||||
&vec![0; n],
|
||||
&mut cache,
|
||||
);
|
||||
cache.truncate_sequence(0, ids.len()).unwrap();
|
||||
}
|
||||
sync();
|
||||
|
||||
// Benchmark single-token decode
|
||||
let mut t = 0.0f64;
|
||||
for i in 0..iters {
|
||||
cache.truncate_sequence(0, ids.len()).unwrap();
|
||||
let t0 = Instant::now();
|
||||
let _ = target.forward_decode_paged(&[ids[0]], &[ids.len()], &[0], &mut cache);
|
||||
sync();
|
||||
t += t0.elapsed().as_secs_f64();
|
||||
let _ = i;
|
||||
}
|
||||
let single = t * 1000.0 / iters as f64;
|
||||
println!(
|
||||
"single-token decode: {:.3} ms (mean of {} iters)",
|
||||
single, iters
|
||||
);
|
||||
|
||||
// Benchmark forward_verify_paged_decode_attention at various batch sizes
|
||||
// (batched-GEMV path).
|
||||
for &n in &[1usize, 2, 3, 5, 9] {
|
||||
let toks: Vec<u32> = (0..n).map(|_| ids[0]).collect();
|
||||
let mut t = 0.0f64;
|
||||
for _ in 0..iters {
|
||||
cache.truncate_sequence(0, ids.len()).unwrap();
|
||||
let t0 = Instant::now();
|
||||
let _ = target.forward_verify_paged_decode_attention(&toks, 0, &mut cache);
|
||||
sync();
|
||||
t += t0.elapsed().as_secs_f64();
|
||||
}
|
||||
let ms = t * 1000.0 / iters as f64;
|
||||
println!(
|
||||
"verify (batched-GEMV) batch={}: {:.3} ms ({:.2}× single)",
|
||||
n,
|
||||
ms,
|
||||
ms / single
|
||||
);
|
||||
}
|
||||
|
||||
// Benchmark _with_hidden variant which uses cuBLAS GEMM after Phase 26 fast-verify.
|
||||
let hooks_layers = [2usize, 18, 33];
|
||||
for &n in &[1usize, 2, 3, 5, 9] {
|
||||
let toks: Vec<u32> = (0..n).map(|_| ids[0]).collect();
|
||||
let mut t = 0.0f64;
|
||||
for _ in 0..iters {
|
||||
cache.truncate_sequence(0, ids.len()).unwrap();
|
||||
let t0 = Instant::now();
|
||||
let _ = target.forward_verify_paged_decode_attention_with_hidden(
|
||||
&toks,
|
||||
0,
|
||||
&mut cache,
|
||||
&hooks_layers,
|
||||
);
|
||||
sync();
|
||||
t += t0.elapsed().as_secs_f64();
|
||||
}
|
||||
let ms = t * 1000.0 / iters as f64;
|
||||
println!(
|
||||
"verify (cuBLAS GEMM) batch={}: {:.3} ms ({:.2}× single)",
|
||||
n,
|
||||
ms,
|
||||
ms / single
|
||||
);
|
||||
}
|
||||
|
||||
cache.free_sequence(0);
|
||||
}
|
||||
|
||||
fn sync() {
|
||||
xserv_cuda::device::synchronize().unwrap();
|
||||
}
|
||||
|
||||
fn arg_usize(args: &[String], flag: &str, default: usize) -> usize {
|
||||
args.iter()
|
||||
.position(|a| a == flag)
|
||||
.and_then(|i| args.get(i + 1))
|
||||
.and_then(|s| s.parse().ok())
|
||||
.unwrap_or(default)
|
||||
}
|
||||
174
crates/xserv-model/src/bin/check-eagle3.rs
Normal file
174
crates/xserv-model/src/bin/check-eagle3.rs
Normal file
@@ -0,0 +1,174 @@
|
||||
//! EAGLE3 sanity check: load weights, run one draft step, print top-5 predictions.
|
||||
//!
|
||||
//! This verifies that:
|
||||
//! - Eagle3Head weights load without shape mismatches
|
||||
//! - Target hidden states can be captured via decode_core_with_hidden
|
||||
//! - Eagle3Head::step produces a valid token id (in target vocab)
|
||||
//!
|
||||
//! Does NOT measure speedup — that requires a full γ≥2 speculative loop, which
|
||||
//! is more complex integration work.
|
||||
|
||||
use std::path::PathBuf;
|
||||
|
||||
use xserv_model::eagle3::{EAGLE_HOOK_LAYERS, Eagle3Head};
|
||||
use xserv_model::{BLOCK_SIZE, ModelConfig, PagedKVCache, Qwen3, loader};
|
||||
use xserv_tensor::{DType, Device, Tensor};
|
||||
use xserv_tokenizer::Tokenizer;
|
||||
|
||||
fn main() {
|
||||
let args: Vec<String> = std::env::args().collect();
|
||||
if args.len() < 3 {
|
||||
eprintln!("Usage: check-eagle3 <target-model-dir> <eagle3-model-dir> [prompt]");
|
||||
std::process::exit(1);
|
||||
}
|
||||
let target_dir = PathBuf::from(&args[1]);
|
||||
let eagle_dir = PathBuf::from(&args[2]);
|
||||
let prompt = args
|
||||
.get(3)
|
||||
.cloned()
|
||||
.unwrap_or_else(|| "The capital of France is".to_string());
|
||||
let device: u32 = 0;
|
||||
|
||||
xserv_cuda::device::set_device(device).unwrap();
|
||||
|
||||
let target_config = ModelConfig::from_file(&target_dir.join("config.json"));
|
||||
eprintln!("Loading target Qwen3-8B...");
|
||||
let target_weights = loader::load_model_dir(&target_dir, Device::Cuda(device));
|
||||
let target = Qwen3::from_weights(target_config.clone(), target_weights);
|
||||
xserv_cuda::allocator::cached_trim();
|
||||
|
||||
eprintln!("Loading EAGLE3 head from {}", eagle_dir.display());
|
||||
let mut eagle = Eagle3Head::load(&eagle_dir, device);
|
||||
xserv_cuda::allocator::cached_trim();
|
||||
|
||||
let tokenizer = Tokenizer::from_file(&target_dir.join("tokenizer.json"));
|
||||
let embed_tokens = target.embed_tokens_tensor();
|
||||
|
||||
let ids = tokenizer.encode(&prompt);
|
||||
let max_seq_len = 512;
|
||||
|
||||
let num_blocks = (max_seq_len + BLOCK_SIZE - 1) / BLOCK_SIZE + 2;
|
||||
let mut cache = PagedKVCache::new(
|
||||
&target_config,
|
||||
num_blocks,
|
||||
0,
|
||||
1,
|
||||
num_blocks,
|
||||
DType::BF16,
|
||||
device,
|
||||
);
|
||||
cache.register_sequence(0).unwrap();
|
||||
|
||||
// Prefill target.
|
||||
let logits = target.forward_prefill_paged(&ids, 0, &mut cache);
|
||||
let target_first = *xserv_kernels::argmax_bf16_to_host(&logits).last().unwrap();
|
||||
let target_first_text = tokenizer.decode(&[target_first]);
|
||||
println!("Prompt: {:?}", prompt);
|
||||
println!(
|
||||
"Target argmax after prefill: {} ({:?})",
|
||||
target_first, target_first_text
|
||||
);
|
||||
|
||||
// Now run one target decode step with target_first to get hidden states at the
|
||||
// hook layers.
|
||||
let pos = cache.seq_len(0);
|
||||
target.decode_prepare(&[pos], &[0], &mut cache);
|
||||
let ids_gpu = upload_u32(&[target_first]);
|
||||
let pos_gpu = upload_u32(&[pos as u32]);
|
||||
let (target_next_logits, hooks) = target.decode_core_with_hidden(
|
||||
ids_gpu.as_ptr() as *const std::ffi::c_void,
|
||||
pos_gpu.as_ptr() as *const std::ffi::c_void,
|
||||
1,
|
||||
&[0],
|
||||
&mut cache,
|
||||
&EAGLE_HOOK_LAYERS,
|
||||
);
|
||||
let target_next = xserv_kernels::argmax_bf16_single(&target_next_logits);
|
||||
let target_next_text = tokenizer.decode(&[target_next]);
|
||||
println!(
|
||||
"Target argmax after 1 decode step: {} ({:?})",
|
||||
target_next, target_next_text
|
||||
);
|
||||
|
||||
for (i, h) in hooks.iter().enumerate() {
|
||||
println!(
|
||||
"hook[{}] (layer {}): shape={:?} dtype={:?}",
|
||||
i,
|
||||
EAGLE_HOOK_LAYERS[i],
|
||||
h.shape(),
|
||||
h.dtype()
|
||||
);
|
||||
}
|
||||
|
||||
// Ask EAGLE what it thinks the NEXT token is (given target_first as prev_token
|
||||
// and the hidden states from the position where target_first lives).
|
||||
// EAGLE should predict target_next (or close to it) to be useful.
|
||||
eagle.reset();
|
||||
let (eagle_pred, eagle_logits) = eagle.step(&hooks, embed_tokens, target_first, pos);
|
||||
let eagle_pred_text = tokenizer.decode(&[eagle_pred]);
|
||||
println!(
|
||||
"EAGLE draft prediction (pairing A: prev=target_first): {} ({:?})",
|
||||
eagle_pred, eagle_pred_text
|
||||
);
|
||||
|
||||
if eagle_pred == target_next {
|
||||
println!("MATCH: EAGLE agrees with target on next token.");
|
||||
} else {
|
||||
println!(
|
||||
"MISMATCH: EAGLE draft={} vs target={} (this is fine per-step; check top-5 below)",
|
||||
eagle_pred, target_next
|
||||
);
|
||||
}
|
||||
|
||||
// Show top-5 from eagle logits (in draft vocab space, mapped to target).
|
||||
print_top5(
|
||||
&eagle_logits,
|
||||
"EAGLE draft top-5 (pairing A)",
|
||||
&eagle,
|
||||
&tokenizer,
|
||||
);
|
||||
|
||||
// Alternative pairing B: pair hooks with target_next (the token those hooks produced
|
||||
// via lm_head), predict token after target_next. Position advances by 1.
|
||||
eagle.reset();
|
||||
let (eagle_pred_b, eagle_logits_b) = eagle.step(&hooks, embed_tokens, target_next, pos + 1);
|
||||
let eagle_pred_b_text = tokenizer.decode(&[eagle_pred_b]);
|
||||
println!(
|
||||
"\nEAGLE draft prediction (pairing B: prev=target_next): {} ({:?})",
|
||||
eagle_pred_b, eagle_pred_b_text
|
||||
);
|
||||
print_top5(
|
||||
&eagle_logits_b,
|
||||
"EAGLE draft top-5 (pairing B)",
|
||||
&eagle,
|
||||
&tokenizer,
|
||||
);
|
||||
}
|
||||
|
||||
fn upload_u32(vals: &[u32]) -> xserv_cuda::GpuBuffer {
|
||||
let bytes = unsafe { std::slice::from_raw_parts(vals.as_ptr() as *const u8, vals.len() * 4) };
|
||||
let mut buf = xserv_cuda::allocator::cached_alloc(bytes.len()).unwrap();
|
||||
buf.copy_from_host(bytes).unwrap();
|
||||
buf
|
||||
}
|
||||
|
||||
fn print_top5(logits: &Tensor, label: &str, eagle: &Eagle3Head, tokenizer: &Tokenizer) {
|
||||
use half::bf16;
|
||||
let cpu = logits.to_device(Device::Cpu);
|
||||
let data = cpu.as_slice::<bf16>();
|
||||
let mut vals: Vec<(usize, f32)> = data
|
||||
.iter()
|
||||
.enumerate()
|
||||
.map(|(i, v)| (i, v.to_f32()))
|
||||
.collect();
|
||||
vals.sort_by(|a, b| b.1.partial_cmp(&a.1).unwrap());
|
||||
println!("{label}:");
|
||||
for (i, val) in vals.iter().take(5) {
|
||||
let target_id = eagle.map_draft_to_target(*i as u32);
|
||||
let text = tokenizer.decode(&[target_id]);
|
||||
println!(
|
||||
" draft_id={} target_id={} val={:.3} text={:?}",
|
||||
i, target_id, val, text
|
||||
);
|
||||
}
|
||||
}
|
||||
425
crates/xserv-model/src/eagle3.rs
Normal file
425
crates/xserv-model/src/eagle3.rs
Normal file
@@ -0,0 +1,425 @@
|
||||
//! EAGLE3 speculative draft head for Qwen3-8B (Phase 25).
|
||||
//!
|
||||
//! Loads the AngelSlim/Qwen3-8B_eagle3 pytorch_model.bin and provides a
|
||||
//! single-step forward pass that takes 3 target hidden states + the previous
|
||||
//! token and returns a draft token in the target vocabulary.
|
||||
//!
|
||||
//! Architecture (from weights):
|
||||
//! - fc: [hidden, 3*hidden] → fuse 3 target hidden states
|
||||
//! - midlayer: 1 decoder layer (attn input dim = 2*hidden)
|
||||
//! - norm + lm_head: → [draft_vocab_size=32000]
|
||||
//! - d2t: draft_id → target_id offset mapping
|
||||
|
||||
use std::collections::HashMap;
|
||||
use std::path::Path;
|
||||
use xserv_kernels::*;
|
||||
use xserv_tensor::{DType, Device, Tensor};
|
||||
|
||||
/// Target layers to hook for EAGLE3 auxiliary hidden states, for Qwen3-8B
|
||||
/// (36 layers). Value comes from AngelSlim/vLLM speculators training config
|
||||
/// `dflash_qwen3_8b_sharegpt_online_5k.sh` which specifies target_layer_ids
|
||||
/// = "2 18 33". Must match training-time selection or EAGLE outputs are wrong.
|
||||
pub const EAGLE_HOOK_LAYERS: [usize; 3] = [2, 18, 33];
|
||||
const DRAFT_VOCAB_SIZE: usize = 32000;
|
||||
|
||||
fn matmul_2d(a: &Tensor, b: &Tensor) -> Tensor {
|
||||
assert_eq!(a.ndim(), 2);
|
||||
assert_eq!(b.ndim(), 2);
|
||||
matmul(a, b, GemmBackend::CuBlas)
|
||||
}
|
||||
|
||||
pub struct Eagle3Head {
|
||||
fc_wt: Tensor, // [hidden, 3*hidden] transposed for matmul
|
||||
hidden_norm: Tensor, // [hidden]
|
||||
input_layernorm: Tensor, // [hidden]
|
||||
q_proj_wt: Tensor, // [num_heads*head_dim, 2*hidden]
|
||||
k_proj_wt: Tensor, // [num_kv_heads*head_dim, 2*hidden]
|
||||
v_proj_wt: Tensor, // [num_kv_heads*head_dim, 2*hidden]
|
||||
o_proj_wt: Tensor, // [hidden, num_heads*head_dim]
|
||||
gate_proj_wt: Tensor, // [intermediate, hidden]
|
||||
up_proj_wt: Tensor, // [intermediate, hidden]
|
||||
down_proj_wt: Tensor, // [hidden, intermediate]
|
||||
post_attention_layernorm: Tensor, // [hidden]
|
||||
norm: Tensor, // [hidden] final
|
||||
lm_head_wt: Tensor, // [draft_vocab, hidden]
|
||||
d2t: Vec<i64>, // [draft_vocab] offset mapping
|
||||
/// t2d[target_id] = true iff target_id has a corresponding draft-vocab id
|
||||
/// (i.e. can potentially be produced by EAGLE). Used to measure the
|
||||
/// coverage cap on acceptance.
|
||||
t2d: Vec<bool>,
|
||||
hidden_size: usize,
|
||||
num_heads: usize,
|
||||
num_kv_heads: usize,
|
||||
head_dim: usize,
|
||||
max_seq_len: usize,
|
||||
rope_cache: RopeCache,
|
||||
// Stateful 1-layer KV cache: [1, num_kv_heads, max_seq_len, head_dim] BF16.
|
||||
// We slice `..current_len` for attention. The head is tiny (~64 KB per
|
||||
// 1000 tokens) so pre-allocating max_seq_len wastes negligible memory.
|
||||
k_cache: Tensor,
|
||||
v_cache: Tensor,
|
||||
current_len: usize,
|
||||
}
|
||||
|
||||
impl Eagle3Head {
|
||||
pub fn load(dir: &Path, device: u32) -> Self {
|
||||
let (weights, d2t, t2d) = load_eagle3_weights(dir, device);
|
||||
let hidden_size = 4096;
|
||||
let num_heads = 32;
|
||||
let num_kv_heads = 8;
|
||||
let head_dim = 128;
|
||||
let intermediate_size = 12288;
|
||||
let max_seq_len = 2048;
|
||||
let rope_theta = 1_000_000.0f32;
|
||||
|
||||
let get = |name: &str| -> Tensor {
|
||||
weights
|
||||
.get(name)
|
||||
.unwrap_or_else(|| panic!("missing eagle3 weight: {name}"))
|
||||
.clone()
|
||||
};
|
||||
|
||||
let fc_wt = get("fc.weight").transpose(0, 1).contiguous();
|
||||
let q_proj_wt = get("midlayer.self_attn.q_proj.weight")
|
||||
.transpose(0, 1)
|
||||
.contiguous();
|
||||
let k_proj_wt = get("midlayer.self_attn.k_proj.weight")
|
||||
.transpose(0, 1)
|
||||
.contiguous();
|
||||
let v_proj_wt = get("midlayer.self_attn.v_proj.weight")
|
||||
.transpose(0, 1)
|
||||
.contiguous();
|
||||
let o_proj_wt = get("midlayer.self_attn.o_proj.weight")
|
||||
.transpose(0, 1)
|
||||
.contiguous();
|
||||
let gate_proj_wt = get("midlayer.mlp.gate_proj.weight")
|
||||
.transpose(0, 1)
|
||||
.contiguous();
|
||||
let up_proj_wt = get("midlayer.mlp.up_proj.weight")
|
||||
.transpose(0, 1)
|
||||
.contiguous();
|
||||
let down_proj_wt = get("midlayer.mlp.down_proj.weight")
|
||||
.transpose(0, 1)
|
||||
.contiguous();
|
||||
let hidden_norm = get("midlayer.hidden_norm.weight");
|
||||
let input_layernorm = get("midlayer.input_layernorm.weight");
|
||||
let post_attention_layernorm = get("midlayer.post_attention_layernorm.weight");
|
||||
let norm = get("norm.weight");
|
||||
let lm_head_wt = get("lm_head.weight").transpose(0, 1).contiguous();
|
||||
|
||||
assert_eq!(d2t.len(), DRAFT_VOCAB_SIZE);
|
||||
|
||||
let rope_cache = RopeCache::new(max_seq_len, head_dim, rope_theta);
|
||||
|
||||
let k_cache = Tensor::zeros(
|
||||
&[1, num_kv_heads, max_seq_len, head_dim],
|
||||
DType::BF16,
|
||||
Device::Cuda(device),
|
||||
);
|
||||
let v_cache = Tensor::zeros(
|
||||
&[1, num_kv_heads, max_seq_len, head_dim],
|
||||
DType::BF16,
|
||||
Device::Cuda(device),
|
||||
);
|
||||
|
||||
Self {
|
||||
fc_wt,
|
||||
hidden_norm,
|
||||
input_layernorm,
|
||||
q_proj_wt,
|
||||
k_proj_wt,
|
||||
v_proj_wt,
|
||||
o_proj_wt,
|
||||
gate_proj_wt,
|
||||
up_proj_wt,
|
||||
down_proj_wt,
|
||||
post_attention_layernorm,
|
||||
norm,
|
||||
lm_head_wt,
|
||||
d2t,
|
||||
t2d,
|
||||
hidden_size,
|
||||
num_heads,
|
||||
num_kv_heads,
|
||||
head_dim,
|
||||
max_seq_len,
|
||||
rope_cache,
|
||||
k_cache,
|
||||
v_cache,
|
||||
current_len: 0,
|
||||
}
|
||||
}
|
||||
|
||||
/// Reset the internal KV cache for a fresh sequence.
|
||||
pub fn reset(&mut self) {
|
||||
self.current_len = 0;
|
||||
}
|
||||
|
||||
/// Truncate the internal KV cache to `new_len` entries. Used to discard
|
||||
/// K/V of rejected drafts after a speculative round.
|
||||
pub fn truncate_to(&mut self, new_len: usize) {
|
||||
assert!(new_len <= self.current_len);
|
||||
self.current_len = new_len;
|
||||
}
|
||||
|
||||
/// Current number of committed K/V entries in the internal EAGLE cache.
|
||||
pub fn current_len(&self) -> usize {
|
||||
self.current_len
|
||||
}
|
||||
|
||||
/// One draft step: produce a token in target vocabulary space.
|
||||
///
|
||||
/// - `target_hidden`: 3 tensors [1, hidden_size] from target hook layers
|
||||
/// - `embed_table`: the target model's embed_tokens (shared, not copied)
|
||||
/// - `prev_token`: the previous committed token
|
||||
/// - `position`: the decode position for RoPE
|
||||
///
|
||||
/// Returns (draft_token_in_target_vocab, draft_logits_tensor).
|
||||
pub fn step(
|
||||
&mut self,
|
||||
target_hidden: &[Tensor; 3],
|
||||
embed_table: &Tensor,
|
||||
prev_token: u32,
|
||||
position: usize,
|
||||
) -> (u32, Tensor) {
|
||||
let (id, logits, _) = self.step_with_aux(target_hidden, embed_table, prev_token, position);
|
||||
(id, logits)
|
||||
}
|
||||
|
||||
/// Like `step`, but also returns the final hidden state (aux) usable as
|
||||
/// the fused_h for a subsequent recursive draft step via `step_recursive`.
|
||||
pub fn step_with_aux(
|
||||
&mut self,
|
||||
target_hidden: &[Tensor; 3],
|
||||
embed_table: &Tensor,
|
||||
prev_token: u32,
|
||||
position: usize,
|
||||
) -> (u32, Tensor, Tensor) {
|
||||
// Fuse 3 target hidden states into fused_h via fc.
|
||||
let h_cat = concat_hidden(target_hidden);
|
||||
let fused_h = matmul_2d(&h_cat, &self.fc_wt);
|
||||
self.forward_from_fused(fused_h, embed_table, prev_token, position)
|
||||
}
|
||||
|
||||
/// Recursive draft step: reuses the previous EAGLE step's aux as fused_h,
|
||||
/// bypassing the fc+3-hidden fusion. Used for γ≥2 chained drafts.
|
||||
pub fn step_recursive(
|
||||
&mut self,
|
||||
fused_h: Tensor,
|
||||
embed_table: &Tensor,
|
||||
prev_token: u32,
|
||||
position: usize,
|
||||
) -> (u32, Tensor, Tensor) {
|
||||
self.forward_from_fused(fused_h, embed_table, prev_token, position)
|
||||
}
|
||||
|
||||
fn forward_from_fused(
|
||||
&mut self,
|
||||
fused_h: Tensor,
|
||||
embed_table: &Tensor,
|
||||
prev_token: u32,
|
||||
position: usize,
|
||||
) -> (u32, Tensor, Tensor) {
|
||||
let eps = 1e-6f32;
|
||||
assert!(
|
||||
self.current_len < self.max_seq_len,
|
||||
"EAGLE KV cache overflow: {} >= {}",
|
||||
self.current_len,
|
||||
self.max_seq_len
|
||||
);
|
||||
|
||||
let emb = embedding(embed_table, &[prev_token]);
|
||||
let residual = fused_h.clone();
|
||||
let emb_normed = rmsnorm(&emb, &self.input_layernorm, eps);
|
||||
let h_normed = rmsnorm(&fused_h, &self.hidden_norm, eps);
|
||||
let attn_in = concat_last_dim(&emb_normed, &h_normed);
|
||||
|
||||
let q = matmul_2d(&attn_in, &self.q_proj_wt);
|
||||
let k = matmul_2d(&attn_in, &self.k_proj_wt);
|
||||
let v = matmul_2d(&attn_in, &self.v_proj_wt);
|
||||
|
||||
let q_3d = q.reshape(&[1, self.num_heads, self.head_dim]);
|
||||
let k_3d = k.reshape(&[1, self.num_kv_heads, self.head_dim]);
|
||||
let positions = [position as u32];
|
||||
rope_inplace(&q_3d, &self.rope_cache, &positions);
|
||||
rope_inplace(&k_3d, &self.rope_cache, &positions);
|
||||
|
||||
let v_3d = v.reshape(&[1, self.num_kv_heads, self.head_dim]);
|
||||
self.append_to_kv_cache(&k_3d, &v_3d);
|
||||
self.current_len += 1;
|
||||
let kv_len = self.current_len;
|
||||
let k_view = self.k_cache.narrow(2, 0, kv_len).contiguous();
|
||||
let v_view = self.v_cache.narrow(2, 0, kv_len).contiguous();
|
||||
|
||||
let q_4d = q_3d.reshape(&[1, self.num_heads, 1, self.head_dim]);
|
||||
let attn_out = decode_attention(&q_4d, &k_view, &v_view);
|
||||
|
||||
let attn_merged = attn_out.reshape(&[1, self.num_heads * self.head_dim]);
|
||||
let attn_proj = matmul_2d(&attn_merged, &self.o_proj_wt);
|
||||
|
||||
let (mlp_in, residual) =
|
||||
add_rmsnorm(&attn_proj, &residual, &self.post_attention_layernorm, eps);
|
||||
|
||||
let gate = matmul_2d(&mlp_in, &self.gate_proj_wt);
|
||||
let up = matmul_2d(&mlp_in, &self.up_proj_wt);
|
||||
let hidden = silu_mul(&gate, &up);
|
||||
let down = matmul_2d(&hidden, &self.down_proj_wt);
|
||||
|
||||
let (x, prenorm) = add_rmsnorm(&down, &residual, &self.norm, eps);
|
||||
let logits = matmul_2d(&x, &self.lm_head_wt);
|
||||
|
||||
let draft_id = argmax_bf16_single(&logits);
|
||||
let target_id = (draft_id as i64 + self.d2t[draft_id as usize]) as u32;
|
||||
// aux for recursive drafting = PRE-norm hidden (default norm_output=False
|
||||
// in vllm/llama_eagle3.py). Feeding the pre-norm state matches training.
|
||||
(target_id, logits, prenorm)
|
||||
}
|
||||
|
||||
/// Write new K/V rows (shape [1, num_kv_heads, head_dim]) at position
|
||||
/// `current_len` inside the [1, num_kv_heads, max_seq_len, head_dim] cache.
|
||||
fn append_to_kv_cache(&mut self, new_k: &Tensor, new_v: &Tensor) {
|
||||
let head_bytes = self.head_dim * self.k_cache.dtype().size_bytes();
|
||||
for h in 0..self.num_kv_heads {
|
||||
for (cache, src) in [(&self.k_cache, new_k), (&self.v_cache, new_v)] {
|
||||
let dst = unsafe {
|
||||
(cache.data_ptr() as *mut u8)
|
||||
.add(((h * self.max_seq_len) + self.current_len) * head_bytes)
|
||||
};
|
||||
let s = unsafe { (src.data_ptr() as *const u8).add(h * head_bytes) };
|
||||
d2d(dst, s, head_bytes);
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
/// Map a draft-vocab token id to the full target-vocab id via d2t.
|
||||
pub fn map_draft_to_target(&self, draft_id: u32) -> u32 {
|
||||
(draft_id as i64 + self.d2t[draft_id as usize]) as u32
|
||||
}
|
||||
|
||||
/// Returns true iff `target_id` is representable in the draft vocabulary
|
||||
/// (i.e., EAGLE could in principle produce it).
|
||||
pub fn target_id_in_draft_vocab(&self, target_id: u32) -> bool {
|
||||
self.t2d.get(target_id as usize).copied().unwrap_or(false)
|
||||
}
|
||||
}
|
||||
|
||||
fn d2d(dst: *mut u8, src: *const u8, bytes: usize) {
|
||||
unsafe {
|
||||
xserv_cuda::ffi::cudaMemcpy(dst, src, bytes, xserv_cuda::ffi::CUDA_MEMCPY_D2D);
|
||||
}
|
||||
}
|
||||
|
||||
fn concat_hidden(hidden: &[Tensor; 3]) -> Tensor {
|
||||
let h = hidden[0].shape()[1];
|
||||
let dtype = hidden[0].dtype();
|
||||
let device = hidden[0].device();
|
||||
let elem_bytes = dtype.size_bytes();
|
||||
let out = Tensor::empty(&[1, 3 * h], dtype, device);
|
||||
for (i, t) in hidden.iter().enumerate() {
|
||||
assert!(t.is_contiguous());
|
||||
let dst = unsafe { (out.data_ptr() as *mut u8).add(i * h * elem_bytes) };
|
||||
d2d(dst, t.data_ptr() as *const u8, h * elem_bytes);
|
||||
}
|
||||
out
|
||||
}
|
||||
|
||||
fn concat_last_dim(a: &Tensor, b: &Tensor) -> Tensor {
|
||||
let da = a.shape()[1];
|
||||
let db = b.shape()[1];
|
||||
let dtype = a.dtype();
|
||||
let device = a.device();
|
||||
let elem_bytes = dtype.size_bytes();
|
||||
let out = Tensor::empty(&[1, da + db], dtype, device);
|
||||
d2d(
|
||||
out.data_ptr() as *mut u8,
|
||||
a.data_ptr() as *const u8,
|
||||
da * elem_bytes,
|
||||
);
|
||||
let dst = unsafe { (out.data_ptr() as *mut u8).add(da * elem_bytes) };
|
||||
d2d(dst, b.data_ptr() as *const u8, db * elem_bytes);
|
||||
out
|
||||
}
|
||||
|
||||
fn repeat_kv_for_single_token(kv: &Tensor, repeats: usize) -> Tensor {
|
||||
if repeats == 1 {
|
||||
return kv.clone();
|
||||
}
|
||||
let nkv = kv.shape()[1];
|
||||
let d = kv.shape()[2];
|
||||
let dtype = kv.dtype();
|
||||
let device = kv.device();
|
||||
let head_bytes = d * dtype.size_bytes();
|
||||
let out = Tensor::empty(&[1, nkv * repeats, d], dtype, device);
|
||||
for h in 0..nkv {
|
||||
let src = unsafe { (kv.data_ptr() as *const u8).add(h * head_bytes) };
|
||||
for r in 0..repeats {
|
||||
let dst = unsafe { (out.data_ptr() as *mut u8).add((h * repeats + r) * head_bytes) };
|
||||
d2d(dst, src, head_bytes);
|
||||
}
|
||||
}
|
||||
out
|
||||
}
|
||||
|
||||
/// Load EAGLE3 weights from safetensors, handling int64 d2t + bool t2d specially.
|
||||
fn load_eagle3_weights(dir: &Path, device: u32) -> (HashMap<String, Tensor>, Vec<i64>, Vec<bool>) {
|
||||
let st_path = dir.join("model.safetensors");
|
||||
assert!(
|
||||
st_path.exists(),
|
||||
"Eagle3 model.safetensors not found in {}. Convert with:\n\
|
||||
python3 -c \"import torch; from safetensors.torch import save_file; \
|
||||
sd=torch.load('pytorch_model.bin', map_location='cpu', weights_only=False); \
|
||||
save_file(sd, 'model.safetensors')\"",
|
||||
dir.display()
|
||||
);
|
||||
|
||||
let data = std::fs::read(&st_path)
|
||||
.unwrap_or_else(|e| panic!("failed to read {}: {e}", st_path.display()));
|
||||
let st = safetensors::SafeTensors::deserialize(&data)
|
||||
.unwrap_or_else(|e| panic!("failed to parse {}: {e}", st_path.display()));
|
||||
|
||||
let mut tensors = HashMap::new();
|
||||
let mut d2t_vec: Vec<i64> = Vec::new();
|
||||
let mut t2d_vec: Vec<bool> = Vec::new();
|
||||
|
||||
for (name, view) in st.tensors() {
|
||||
if name == "t2d" {
|
||||
let raw = view.data();
|
||||
assert_eq!(view.dtype(), safetensors::Dtype::BOOL);
|
||||
t2d_vec = raw.iter().map(|&b| b != 0).collect();
|
||||
continue;
|
||||
}
|
||||
if name == "d2t" {
|
||||
let raw = view.data();
|
||||
assert_eq!(view.dtype(), safetensors::Dtype::I64);
|
||||
let n = raw.len() / 8;
|
||||
d2t_vec = (0..n)
|
||||
.map(|i| i64::from_le_bytes(raw[i * 8..(i + 1) * 8].try_into().unwrap()))
|
||||
.collect();
|
||||
continue;
|
||||
}
|
||||
let dtype = match view.dtype() {
|
||||
safetensors::Dtype::BF16 => DType::BF16,
|
||||
safetensors::Dtype::F32 => DType::F32,
|
||||
safetensors::Dtype::F16 => DType::F16,
|
||||
other => {
|
||||
eprintln!("eagle3: skipping {name} with unsupported dtype {other:?}");
|
||||
continue;
|
||||
}
|
||||
};
|
||||
let shape: Vec<usize> = view.shape().to_vec();
|
||||
let raw = view.data();
|
||||
let t = crate::loader::make_tensor(raw, &shape, dtype);
|
||||
let t = t.to_device(Device::Cuda(device));
|
||||
tensors.insert(name.to_string(), t);
|
||||
}
|
||||
|
||||
assert!(
|
||||
!d2t_vec.is_empty(),
|
||||
"d2t tensor not found in eagle3 weights"
|
||||
);
|
||||
assert!(
|
||||
!t2d_vec.is_empty(),
|
||||
"t2d tensor not found in eagle3 weights"
|
||||
);
|
||||
(tensors, d2t_vec, t2d_vec)
|
||||
}
|
||||
@@ -1,5 +1,6 @@
|
||||
pub mod config;
|
||||
pub mod decode_graph;
|
||||
pub mod eagle3;
|
||||
pub mod gpt2;
|
||||
pub mod gpt_oss;
|
||||
pub mod gpt_oss_graph;
|
||||
@@ -7,6 +8,7 @@ pub mod kv_cache;
|
||||
pub mod loader;
|
||||
pub mod paged_kv_cache;
|
||||
pub mod qwen3;
|
||||
pub mod qwen3_graph;
|
||||
pub mod sampling;
|
||||
|
||||
pub use config::ModelConfig;
|
||||
|
||||
@@ -68,7 +68,7 @@ pub fn load_model_dir(dir: &Path, device: Device) -> HashMap<String, Tensor> {
|
||||
all_tensors
|
||||
}
|
||||
|
||||
fn make_tensor(raw_bytes: &[u8], shape: &[usize], dtype: DType) -> Tensor {
|
||||
pub(crate) fn make_tensor(raw_bytes: &[u8], shape: &[usize], dtype: DType) -> Tensor {
|
||||
match dtype {
|
||||
DType::F32 => {
|
||||
let floats: &[f32] = unsafe {
|
||||
|
||||
@@ -515,6 +515,51 @@ impl PagedKVCache {
|
||||
Ok(())
|
||||
}
|
||||
|
||||
/// Copy K/V data from `src_pos` to `dst_pos` within the same slot, across
|
||||
/// all layers. Used by tree speculative decoding to remap an accepted
|
||||
/// sibling's K/V to the canonical sequential position after acceptance.
|
||||
///
|
||||
/// Requires: both positions within the currently-allocated block range.
|
||||
pub fn copy_kv_position(&self, slot: usize, src_pos: usize, dst_pos: usize) {
|
||||
let state = self.seq_states[slot]
|
||||
.as_ref()
|
||||
.expect("copy_kv_position: slot not registered");
|
||||
assert!(
|
||||
src_pos < state.seq_len && dst_pos < state.seq_len,
|
||||
"copy_kv_position: positions must be within seq_len"
|
||||
);
|
||||
// Upload this sequence's block_ids to a small GPU buffer.
|
||||
let block_ids_host: Vec<i32> = state.block_ids.iter().map(|&b| b as i32).collect();
|
||||
let bytes: &[u8] = unsafe {
|
||||
std::slice::from_raw_parts(
|
||||
block_ids_host.as_ptr() as *const u8,
|
||||
block_ids_host.len() * 4,
|
||||
)
|
||||
};
|
||||
let mut ids_buf =
|
||||
xserv_cuda::allocator::cached_alloc(bytes.len()).expect("alloc block_ids for copy");
|
||||
ids_buf.copy_from_host(bytes).unwrap();
|
||||
let ids_ptr = ids_buf.as_ptr() as *const i32;
|
||||
|
||||
let stream = xserv_cuda::current_stream_raw();
|
||||
let num_layers = self.k_pools.len();
|
||||
for layer in 0..num_layers {
|
||||
unsafe {
|
||||
xserv_kernels::copy_kv_position(
|
||||
self.k_pools[layer].as_ptr() as *mut std::ffi::c_void,
|
||||
self.v_pools[layer].as_ptr() as *mut std::ffi::c_void,
|
||||
ids_ptr,
|
||||
src_pos,
|
||||
dst_pos,
|
||||
self.num_kv_heads,
|
||||
self.head_dim,
|
||||
BLOCK_SIZE,
|
||||
stream,
|
||||
);
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
/// Refresh the host-side block table + context lens from `seq_states`,
|
||||
/// then upload to GPU. Call once per decode step before the paged kernel.
|
||||
pub fn sync_to_gpu(&mut self) {
|
||||
|
||||
@@ -701,45 +701,72 @@ impl Qwen3 {
|
||||
assert_eq!(seq_slots.len(), batch);
|
||||
assert!(batch > 0);
|
||||
|
||||
// TP: this rank owns a slice of the heads (local_* == full when world==1).
|
||||
let num_heads = self.local_num_heads;
|
||||
let num_kv_heads = self.local_num_kv_heads;
|
||||
let head_dim = self.config.head_dim();
|
||||
let eps = self.config.rms_norm_eps.unwrap_or(1e-6) as f32;
|
||||
self.decode_prepare(positions, seq_slots, paged_cache);
|
||||
|
||||
// Ensure all slots have enough physical blocks for this token, then
|
||||
// upload block tables + context_lens once for the whole forward (the
|
||||
// tables are identical across layers; only the layer's K/V pool changes).
|
||||
let ids_gpu = upload_u32(tokens);
|
||||
let positions_u32: Vec<u32> = positions.iter().map(|&p| p as u32).collect();
|
||||
let pos_gpu = upload_u32(&positions_u32);
|
||||
let logits = self.decode_core(
|
||||
ids_gpu.as_ptr() as *const std::ffi::c_void,
|
||||
pos_gpu.as_ptr() as *const std::ffi::c_void,
|
||||
batch,
|
||||
seq_slots,
|
||||
paged_cache,
|
||||
);
|
||||
logits
|
||||
}
|
||||
|
||||
/// Host-side per-step cache bookkeeping: block allocation + uploading block
|
||||
/// tables / context lens to their (stable-address) GPU buffers. Runs
|
||||
/// OUTSIDE any CUDA-graph captured region.
|
||||
pub fn decode_prepare(
|
||||
&self,
|
||||
positions: &[usize],
|
||||
seq_slots: &[usize],
|
||||
paged_cache: &mut PagedKVCache,
|
||||
) {
|
||||
let kv_lens: Vec<i32> = positions.iter().map(|&p| (p + 1) as i32).collect();
|
||||
for (b, &slot) in seq_slots.iter().enumerate() {
|
||||
paged_cache.ensure_capacity(slot, positions[b] + 1);
|
||||
}
|
||||
paged_cache.sync_active_batch_with_lens(seq_slots, &kv_lens);
|
||||
}
|
||||
|
||||
/// Pure-GPU decode step: embedding → all layers → final norm → logits.
|
||||
/// Token ids and positions are read from device buffers; every other input
|
||||
/// (weights, KV pools, block table, context lens) has a stable address —
|
||||
/// which makes this region CUDA-graph capturable.
|
||||
pub fn decode_core(
|
||||
&self,
|
||||
ids_gpu: *const std::ffi::c_void,
|
||||
pos_gpu: *const std::ffi::c_void,
|
||||
batch: usize,
|
||||
seq_slots: &[usize],
|
||||
paged_cache: &mut PagedKVCache,
|
||||
) -> Tensor {
|
||||
let num_heads = self.local_num_heads;
|
||||
let num_kv_heads = self.local_num_kv_heads;
|
||||
let head_dim = self.config.head_dim();
|
||||
let eps = self.config.rms_norm_eps.unwrap_or(1e-6) as f32;
|
||||
|
||||
let bt_ptr = paged_cache.block_table_gpu().as_ptr() as *const i32;
|
||||
let cl_ptr = paged_cache.context_lens_gpu().as_ptr() as *const i32;
|
||||
let max_blocks = paged_cache.max_blocks_per_seq();
|
||||
|
||||
// RoPE expects `[num_tokens, H, D]` with `num_tokens` positions —
|
||||
// matches our `[B, H, D]` exactly, so we upload once here.
|
||||
let positions_u32: Vec<u32> = positions.iter().map(|&p| p as u32).collect();
|
||||
|
||||
// Batched embedding: [B, hidden]
|
||||
let mut x = embedding(&self.embed_tokens, tokens);
|
||||
let mut x = embedding_device_ids(&self.embed_tokens, ids_gpu, batch);
|
||||
|
||||
for (layer_idx, layer) in self.layers.iter().enumerate() {
|
||||
let residual = x.clone();
|
||||
let normed = rmsnorm(&x, &layer.input_norm, eps);
|
||||
|
||||
// Fused QKV projection: one GEMV instead of three.
|
||||
let qkv = matmul_2d(&normed, &layer.qkv_proj_wt); // [B, (H+2*KV)*D]
|
||||
let qkv = matmul_2d(&normed, &layer.qkv_proj_wt);
|
||||
let q_dim = num_heads * head_dim;
|
||||
let kv_dim = num_kv_heads * head_dim;
|
||||
let q_all = qkv.narrow(1, 0, q_dim); // [B, H*D] (view)
|
||||
let k_all = qkv.narrow(1, q_dim, kv_dim); // [B, KV*D] (view)
|
||||
let q_all = qkv.narrow(1, 0, q_dim);
|
||||
let k_all = qkv.narrow(1, q_dim, kv_dim);
|
||||
let v_all = qkv.narrow(1, q_dim + kv_dim, kv_dim);
|
||||
|
||||
// Per-head RMSNorm on contiguous copies (narrow views are strided).
|
||||
let q_flat = q_all.contiguous().reshape(&[batch * num_heads, head_dim]);
|
||||
let k_flat = k_all
|
||||
.contiguous()
|
||||
@@ -749,16 +776,13 @@ impl Qwen3 {
|
||||
|
||||
let q_3d = q_normed.reshape(&[batch, num_heads, head_dim]);
|
||||
let k_3d = k_normed.reshape(&[batch, num_kv_heads, head_dim]);
|
||||
rope_inplace(&q_3d, &self.rope_cache, &positions_u32);
|
||||
rope_inplace(&k_3d, &self.rope_cache, &positions_u32);
|
||||
rope_inplace_device_pos(&q_3d, &self.rope_cache, pos_gpu);
|
||||
rope_inplace_device_pos(&k_3d, &self.rope_cache, pos_gpu);
|
||||
|
||||
let v_3d = v_all.contiguous().reshape(&[batch, num_kv_heads, head_dim]);
|
||||
|
||||
// Single batched scatter for all sequences in the batch.
|
||||
paged_cache.append_tokens_batched(layer_idx, &k_3d, &v_3d, batch);
|
||||
|
||||
// Paged attention reads Q as [B, H, 1, D] — a contiguous view
|
||||
// of [B, H, D].
|
||||
let q_4d = q_3d.reshape(&[batch, num_heads, 1, head_dim]);
|
||||
let k_pool_ptr = paged_cache.k_pool(layer_idx).as_ptr() as *const std::ffi::c_void;
|
||||
let v_pool_ptr = paged_cache.v_pool(layer_idx).as_ptr() as *const std::ffi::c_void;
|
||||
@@ -775,27 +799,24 @@ impl Qwen3 {
|
||||
max_blocks,
|
||||
);
|
||||
|
||||
// attn_out shape [B, H, 1, D] is contiguous-equivalent to [B, H*D].
|
||||
let attn_merged = attn_out.reshape(&[batch, num_heads * head_dim]);
|
||||
let attn_proj = matmul_2d(&attn_merged, &layer.o_proj_wt);
|
||||
self.all_reduce(&attn_proj); // TP: sum partial attention outputs
|
||||
self.all_reduce(&attn_proj);
|
||||
|
||||
let (normed, x_new) =
|
||||
xserv_kernels::add_rmsnorm(&attn_proj, &residual, &layer.post_norm, eps);
|
||||
let residual = x_new.clone();
|
||||
|
||||
// Fused gate+up projection: one GEMV instead of two.
|
||||
let gate_up = matmul_2d(&normed, &layer.gate_up_proj_wt); // [B, 2*ffn]
|
||||
let gate_up = matmul_2d(&normed, &layer.gate_up_proj_wt);
|
||||
let ffn_dim = gate_up.shape()[1] / 2;
|
||||
let gate = gate_up.narrow(1, 0, ffn_dim).contiguous();
|
||||
let up = gate_up.narrow(1, ffn_dim, ffn_dim).contiguous();
|
||||
let hidden_states = xserv_kernels::silu_mul(&gate, &up);
|
||||
let down = matmul_2d(&hidden_states, &layer.down_proj_wt);
|
||||
self.all_reduce(&down); // TP: sum partial MLP outputs
|
||||
self.all_reduce(&down);
|
||||
x = add_any(&residual, &down);
|
||||
}
|
||||
|
||||
// Advance logical seq_len now that all layers have been written.
|
||||
for &slot in seq_slots {
|
||||
paged_cache.advance_seq_len(slot, 1);
|
||||
}
|
||||
@@ -804,6 +825,111 @@ impl Qwen3 {
|
||||
matmul_2d(&x, &self.lm_head_t)
|
||||
}
|
||||
|
||||
/// Like `decode_core` but also captures hidden states at 3 specified layer
|
||||
/// indices (after residual+MLP output). Used by EAGLE3 speculative drafting
|
||||
/// to feed the draft head with low/mid/high target representations.
|
||||
pub fn decode_core_with_hidden(
|
||||
&self,
|
||||
ids_gpu: *const std::ffi::c_void,
|
||||
pos_gpu: *const std::ffi::c_void,
|
||||
batch: usize,
|
||||
seq_slots: &[usize],
|
||||
paged_cache: &mut PagedKVCache,
|
||||
hook_layers: &[usize; 3],
|
||||
) -> (Tensor, [Tensor; 3]) {
|
||||
let num_heads = self.local_num_heads;
|
||||
let num_kv_heads = self.local_num_kv_heads;
|
||||
let head_dim = self.config.head_dim();
|
||||
let eps = self.config.rms_norm_eps.unwrap_or(1e-6) as f32;
|
||||
|
||||
let bt_ptr = paged_cache.block_table_gpu().as_ptr() as *const i32;
|
||||
let cl_ptr = paged_cache.context_lens_gpu().as_ptr() as *const i32;
|
||||
let max_blocks = paged_cache.max_blocks_per_seq();
|
||||
|
||||
let mut x = embedding_device_ids(&self.embed_tokens, ids_gpu, batch);
|
||||
let mut hooks: [Option<Tensor>; 3] = [None, None, None];
|
||||
|
||||
for (layer_idx, layer) in self.layers.iter().enumerate() {
|
||||
let residual = x.clone();
|
||||
let normed = rmsnorm(&x, &layer.input_norm, eps);
|
||||
|
||||
let qkv = matmul_2d(&normed, &layer.qkv_proj_wt);
|
||||
let q_dim = num_heads * head_dim;
|
||||
let kv_dim = num_kv_heads * head_dim;
|
||||
let q_all = qkv.narrow(1, 0, q_dim);
|
||||
let k_all = qkv.narrow(1, q_dim, kv_dim);
|
||||
let v_all = qkv.narrow(1, q_dim + kv_dim, kv_dim);
|
||||
|
||||
let q_flat = q_all.contiguous().reshape(&[batch * num_heads, head_dim]);
|
||||
let k_flat = k_all
|
||||
.contiguous()
|
||||
.reshape(&[batch * num_kv_heads, head_dim]);
|
||||
let q_normed = rmsnorm(&q_flat, &layer.q_norm, eps);
|
||||
let k_normed = rmsnorm(&k_flat, &layer.k_norm, eps);
|
||||
|
||||
let q_3d = q_normed.reshape(&[batch, num_heads, head_dim]);
|
||||
let k_3d = k_normed.reshape(&[batch, num_kv_heads, head_dim]);
|
||||
rope_inplace_device_pos(&q_3d, &self.rope_cache, pos_gpu);
|
||||
rope_inplace_device_pos(&k_3d, &self.rope_cache, pos_gpu);
|
||||
|
||||
let v_3d = v_all.contiguous().reshape(&[batch, num_kv_heads, head_dim]);
|
||||
|
||||
paged_cache.append_tokens_batched(layer_idx, &k_3d, &v_3d, batch);
|
||||
|
||||
let q_4d = q_3d.reshape(&[batch, num_heads, 1, head_dim]);
|
||||
let k_pool_ptr = paged_cache.k_pool(layer_idx).as_ptr() as *const std::ffi::c_void;
|
||||
let v_pool_ptr = paged_cache.v_pool(layer_idx).as_ptr() as *const std::ffi::c_void;
|
||||
let attn_out = xserv_kernels::paged_decode_attention(
|
||||
&q_4d,
|
||||
k_pool_ptr,
|
||||
v_pool_ptr,
|
||||
bt_ptr,
|
||||
cl_ptr,
|
||||
batch,
|
||||
num_heads,
|
||||
num_kv_heads,
|
||||
head_dim,
|
||||
max_blocks,
|
||||
);
|
||||
|
||||
let attn_merged = attn_out.reshape(&[batch, num_heads * head_dim]);
|
||||
let attn_proj = matmul_2d(&attn_merged, &layer.o_proj_wt);
|
||||
self.all_reduce(&attn_proj);
|
||||
|
||||
let (normed, x_new) =
|
||||
xserv_kernels::add_rmsnorm(&attn_proj, &residual, &layer.post_norm, eps);
|
||||
let residual = x_new.clone();
|
||||
|
||||
let gate_up = matmul_2d(&normed, &layer.gate_up_proj_wt);
|
||||
let ffn_dim = gate_up.shape()[1] / 2;
|
||||
let gate = gate_up.narrow(1, 0, ffn_dim).contiguous();
|
||||
let up = gate_up.narrow(1, ffn_dim, ffn_dim).contiguous();
|
||||
let hidden_states = xserv_kernels::silu_mul(&gate, &up);
|
||||
let down = matmul_2d(&hidden_states, &layer.down_proj_wt);
|
||||
self.all_reduce(&down);
|
||||
x = add_any(&residual, &down);
|
||||
|
||||
for (h_idx, &h_layer) in hook_layers.iter().enumerate() {
|
||||
if layer_idx == h_layer {
|
||||
hooks[h_idx] = Some(x.clone());
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
for &slot in seq_slots {
|
||||
paged_cache.advance_seq_len(slot, 1);
|
||||
}
|
||||
|
||||
let x = rmsnorm(&x, &self.norm, eps);
|
||||
let logits = matmul_2d(&x, &self.lm_head_t);
|
||||
let hidden_arr = [
|
||||
hooks[0].take().expect("hook layer 0 not reached"),
|
||||
hooks[1].take().expect("hook layer 1 not reached"),
|
||||
hooks[2].take().expect("hook layer 2 not reached"),
|
||||
];
|
||||
(logits, hidden_arr)
|
||||
}
|
||||
|
||||
/// Paged prefill: write a sequence of `new_tokens` K/V into the paged
|
||||
/// cache for `slot`, run flash attention via gathered contiguous K/V.
|
||||
/// Returns logits [new_tokens, vocab_size].
|
||||
@@ -923,7 +1049,7 @@ impl Qwen3 {
|
||||
let residual = x.clone();
|
||||
let normed = rmsnorm(&x, &layer.input_norm, eps);
|
||||
|
||||
let qkv = matmul_rows_gemv(&normed, &layer.qkv_proj_wt);
|
||||
let qkv = matmul_batched_gemv(&normed, &layer.qkv_proj_wt);
|
||||
let q_dim = num_heads * head_dim;
|
||||
let kv_dim = num_kv_heads * head_dim;
|
||||
let q_all = qkv.narrow(1, 0, q_dim);
|
||||
@@ -966,25 +1092,274 @@ impl Qwen3 {
|
||||
);
|
||||
|
||||
let attn_merged = attn_out.reshape(&[new_tokens, num_heads * head_dim]);
|
||||
let attn_proj = matmul_rows_gemv(&attn_merged, &layer.o_proj_wt);
|
||||
let attn_proj = matmul_batched_gemv(&attn_merged, &layer.o_proj_wt);
|
||||
self.all_reduce(&attn_proj);
|
||||
|
||||
let (normed, x_new) =
|
||||
xserv_kernels::add_rmsnorm(&attn_proj, &residual, &layer.post_norm, eps);
|
||||
let residual = x_new.clone();
|
||||
|
||||
let gate_up = matmul_rows_gemv(&normed, &layer.gate_up_proj_wt);
|
||||
let gate_up = matmul_batched_gemv(&normed, &layer.gate_up_proj_wt);
|
||||
let ffn_dim = gate_up.shape()[1] / 2;
|
||||
let gate = gate_up.narrow(1, 0, ffn_dim).contiguous();
|
||||
let up = gate_up.narrow(1, ffn_dim, ffn_dim).contiguous();
|
||||
let hidden_states = xserv_kernels::silu_mul(&gate, &up);
|
||||
let down = matmul_rows_gemv(&hidden_states, &layer.down_proj_wt);
|
||||
let down = matmul_batched_gemv(&hidden_states, &layer.down_proj_wt);
|
||||
self.all_reduce(&down);
|
||||
x = add_any(&residual, &down);
|
||||
}
|
||||
|
||||
let x = rmsnorm(&x, &self.norm, eps);
|
||||
matmul_rows_gemv(&x, &self.lm_head_t)
|
||||
matmul_batched_gemv(&x, &self.lm_head_t)
|
||||
}
|
||||
|
||||
/// Like `forward_verify_paged_decode_attention`, but also captures hidden
|
||||
/// states at 3 layer indices (per position). Returns
|
||||
/// (logits [new_tokens, vocab], hooks [3][new_tokens, hidden]). Used by
|
||||
/// EAGLE3 speculative γ≥2 verify path so we can seed the next round's
|
||||
/// EAGLE draft with target's real hidden states at the accepted position.
|
||||
pub fn forward_verify_paged_decode_attention_with_hidden(
|
||||
&self,
|
||||
token_ids: &[u32],
|
||||
slot: usize,
|
||||
paged_cache: &mut PagedKVCache,
|
||||
hook_layers: &[usize; 3],
|
||||
) -> (Tensor, [Tensor; 3]) {
|
||||
let new_tokens = token_ids.len();
|
||||
let pos_offset = paged_cache.seq_len(slot);
|
||||
let num_heads = self.local_num_heads;
|
||||
let num_kv_heads = self.local_num_kv_heads;
|
||||
let head_dim = self.config.head_dim();
|
||||
let eps = self.config.rms_norm_eps.unwrap_or(1e-6) as f32;
|
||||
|
||||
paged_cache.ensure_capacity(slot, pos_offset + new_tokens);
|
||||
paged_cache.advance_seq_len(slot, new_tokens);
|
||||
|
||||
let positions: Vec<u32> = (pos_offset..pos_offset + new_tokens)
|
||||
.map(|p| p as u32)
|
||||
.collect();
|
||||
let kv_lens: Vec<i32> = (0..new_tokens)
|
||||
.map(|i| (pos_offset + i + 1) as i32)
|
||||
.collect();
|
||||
let slots = vec![slot; new_tokens];
|
||||
paged_cache.sync_active_batch_with_lens(&slots, &kv_lens);
|
||||
let bt_ptr = paged_cache.block_table_gpu().as_ptr() as *const i32;
|
||||
let cl_ptr = paged_cache.context_lens_gpu().as_ptr() as *const i32;
|
||||
let max_blocks = paged_cache.max_blocks_per_seq();
|
||||
|
||||
let mut x = embedding(&self.embed_tokens, token_ids);
|
||||
let mut hooks: [Option<Tensor>; 3] = [None, None, None];
|
||||
|
||||
for (layer_idx, layer) in self.layers.iter().enumerate() {
|
||||
let residual = x.clone();
|
||||
let normed = rmsnorm(&x, &layer.input_norm, eps);
|
||||
|
||||
let qkv = matmul_2d(&normed, &layer.qkv_proj_wt);
|
||||
let q_dim = num_heads * head_dim;
|
||||
let kv_dim = num_kv_heads * head_dim;
|
||||
let q_all = qkv.narrow(1, 0, q_dim);
|
||||
let k_all = qkv.narrow(1, q_dim, kv_dim);
|
||||
let v_all = qkv.narrow(1, q_dim + kv_dim, kv_dim);
|
||||
|
||||
let q_flat = q_all
|
||||
.contiguous()
|
||||
.reshape(&[new_tokens * num_heads, head_dim]);
|
||||
let k_flat = k_all
|
||||
.contiguous()
|
||||
.reshape(&[new_tokens * num_kv_heads, head_dim]);
|
||||
let q_normed = rmsnorm(&q_flat, &layer.q_norm, eps);
|
||||
let k_normed = rmsnorm(&k_flat, &layer.k_norm, eps);
|
||||
|
||||
let q_3d = q_normed.reshape(&[new_tokens, num_heads, head_dim]);
|
||||
let k_3d = k_normed.reshape(&[new_tokens, num_kv_heads, head_dim]);
|
||||
rope_inplace(&q_3d, &self.rope_cache, &positions);
|
||||
rope_inplace(&k_3d, &self.rope_cache, &positions);
|
||||
|
||||
let v_3d = v_all
|
||||
.contiguous()
|
||||
.reshape(&[new_tokens, num_kv_heads, head_dim]);
|
||||
paged_cache.append_tokens_batched(layer_idx, &k_3d, &v_3d, new_tokens);
|
||||
|
||||
let q_decode = q_3d.reshape(&[new_tokens, num_heads, 1, head_dim]);
|
||||
let k_pool_ptr = paged_cache.k_pool(layer_idx).as_ptr() as *const std::ffi::c_void;
|
||||
let v_pool_ptr = paged_cache.v_pool(layer_idx).as_ptr() as *const std::ffi::c_void;
|
||||
let attn_out = xserv_kernels::paged_decode_attention(
|
||||
&q_decode,
|
||||
k_pool_ptr,
|
||||
v_pool_ptr,
|
||||
bt_ptr,
|
||||
cl_ptr,
|
||||
new_tokens,
|
||||
num_heads,
|
||||
num_kv_heads,
|
||||
head_dim,
|
||||
max_blocks,
|
||||
);
|
||||
|
||||
let attn_merged = attn_out.reshape(&[new_tokens, num_heads * head_dim]);
|
||||
let attn_proj = matmul_2d(&attn_merged, &layer.o_proj_wt);
|
||||
self.all_reduce(&attn_proj);
|
||||
|
||||
let (normed, x_new) =
|
||||
xserv_kernels::add_rmsnorm(&attn_proj, &residual, &layer.post_norm, eps);
|
||||
let residual = x_new.clone();
|
||||
|
||||
let gate_up = matmul_2d(&normed, &layer.gate_up_proj_wt);
|
||||
let ffn_dim = gate_up.shape()[1] / 2;
|
||||
let gate = gate_up.narrow(1, 0, ffn_dim).contiguous();
|
||||
let up = gate_up.narrow(1, ffn_dim, ffn_dim).contiguous();
|
||||
let hidden_states = xserv_kernels::silu_mul(&gate, &up);
|
||||
let down = matmul_2d(&hidden_states, &layer.down_proj_wt);
|
||||
self.all_reduce(&down);
|
||||
x = add_any(&residual, &down);
|
||||
|
||||
for (h_idx, &h_layer) in hook_layers.iter().enumerate() {
|
||||
if layer_idx == h_layer {
|
||||
hooks[h_idx] = Some(x.clone());
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
let x = rmsnorm(&x, &self.norm, eps);
|
||||
let logits = matmul_2d(&x, &self.lm_head_t);
|
||||
let hidden_arr = [
|
||||
hooks[0].take().expect("hook layer 0 not reached"),
|
||||
hooks[1].take().expect("hook layer 1 not reached"),
|
||||
hooks[2].take().expect("hook layer 2 not reached"),
|
||||
];
|
||||
(logits, hidden_arr)
|
||||
}
|
||||
|
||||
/// Tree-aware verify: like `_with_hidden` but supports sibling candidates
|
||||
/// sharing the same target position. Caller supplies per-token positions
|
||||
/// (for RoPE), kv_lens (attention context length), and a flattened
|
||||
/// `tree_mask` (`[new_tokens, new_tokens]` i32; `mask[i, j]!=0` iff query i
|
||||
/// attends to newly-written K/V at slot j). Positions in the paged cache
|
||||
/// before pos_offset are always attended (regular history).
|
||||
#[allow(clippy::too_many_arguments)]
|
||||
pub fn forward_verify_paged_decode_attention_tree_with_hidden(
|
||||
&self,
|
||||
token_ids: &[u32],
|
||||
positions: &[u32],
|
||||
kv_lens: &[i32],
|
||||
tree_mask: &[i32],
|
||||
slot: usize,
|
||||
paged_cache: &mut PagedKVCache,
|
||||
hook_layers: &[usize; 3],
|
||||
) -> (Tensor, [Tensor; 3]) {
|
||||
let new_tokens = token_ids.len();
|
||||
assert_eq!(positions.len(), new_tokens);
|
||||
assert_eq!(kv_lens.len(), new_tokens);
|
||||
assert_eq!(tree_mask.len(), new_tokens * new_tokens);
|
||||
|
||||
let pos_offset = paged_cache.seq_len(slot);
|
||||
let num_heads = self.local_num_heads;
|
||||
let num_kv_heads = self.local_num_kv_heads;
|
||||
let head_dim = self.config.head_dim();
|
||||
let eps = self.config.rms_norm_eps.unwrap_or(1e-6) as f32;
|
||||
|
||||
paged_cache.ensure_capacity(slot, pos_offset + new_tokens);
|
||||
paged_cache.advance_seq_len(slot, new_tokens);
|
||||
|
||||
let slots = vec![slot; new_tokens];
|
||||
paged_cache.sync_active_batch_with_lens(&slots, kv_lens);
|
||||
let bt_ptr = paged_cache.block_table_gpu().as_ptr() as *const i32;
|
||||
let cl_ptr = paged_cache.context_lens_gpu().as_ptr() as *const i32;
|
||||
let max_blocks = paged_cache.max_blocks_per_seq();
|
||||
|
||||
// Upload tree_mask [new_tokens, new_tokens] i32 to GPU.
|
||||
let mask_bytes: &[u8] = unsafe {
|
||||
std::slice::from_raw_parts(tree_mask.as_ptr() as *const u8, tree_mask.len() * 4)
|
||||
};
|
||||
let mut mask_buf =
|
||||
xserv_cuda::allocator::cached_alloc(mask_bytes.len()).expect("alloc tree_mask");
|
||||
mask_buf.copy_from_host(mask_bytes).unwrap();
|
||||
let mask_ptr = mask_buf.as_ptr() as *const i32;
|
||||
|
||||
let mut x = embedding(&self.embed_tokens, token_ids);
|
||||
let mut hooks: [Option<Tensor>; 3] = [None, None, None];
|
||||
|
||||
for (layer_idx, layer) in self.layers.iter().enumerate() {
|
||||
let residual = x.clone();
|
||||
let normed = rmsnorm(&x, &layer.input_norm, eps);
|
||||
|
||||
let qkv = matmul_2d(&normed, &layer.qkv_proj_wt);
|
||||
let q_dim = num_heads * head_dim;
|
||||
let kv_dim = num_kv_heads * head_dim;
|
||||
let q_all = qkv.narrow(1, 0, q_dim);
|
||||
let k_all = qkv.narrow(1, q_dim, kv_dim);
|
||||
let v_all = qkv.narrow(1, q_dim + kv_dim, kv_dim);
|
||||
|
||||
let q_flat = q_all
|
||||
.contiguous()
|
||||
.reshape(&[new_tokens * num_heads, head_dim]);
|
||||
let k_flat = k_all
|
||||
.contiguous()
|
||||
.reshape(&[new_tokens * num_kv_heads, head_dim]);
|
||||
let q_normed = rmsnorm(&q_flat, &layer.q_norm, eps);
|
||||
let k_normed = rmsnorm(&k_flat, &layer.k_norm, eps);
|
||||
|
||||
let q_3d = q_normed.reshape(&[new_tokens, num_heads, head_dim]);
|
||||
let k_3d = k_normed.reshape(&[new_tokens, num_kv_heads, head_dim]);
|
||||
rope_inplace(&q_3d, &self.rope_cache, positions);
|
||||
rope_inplace(&k_3d, &self.rope_cache, positions);
|
||||
|
||||
let v_3d = v_all
|
||||
.contiguous()
|
||||
.reshape(&[new_tokens, num_kv_heads, head_dim]);
|
||||
paged_cache.append_tokens_batched(layer_idx, &k_3d, &v_3d, new_tokens);
|
||||
|
||||
let q_decode = q_3d.reshape(&[new_tokens, num_heads, 1, head_dim]);
|
||||
let k_pool_ptr = paged_cache.k_pool(layer_idx).as_ptr() as *const std::ffi::c_void;
|
||||
let v_pool_ptr = paged_cache.v_pool(layer_idx).as_ptr() as *const std::ffi::c_void;
|
||||
let attn_out = xserv_kernels::paged_decode_attention_tree(
|
||||
&q_decode,
|
||||
k_pool_ptr,
|
||||
v_pool_ptr,
|
||||
bt_ptr,
|
||||
cl_ptr,
|
||||
mask_ptr,
|
||||
new_tokens,
|
||||
num_heads,
|
||||
num_kv_heads,
|
||||
head_dim,
|
||||
max_blocks,
|
||||
pos_offset,
|
||||
new_tokens,
|
||||
);
|
||||
|
||||
let attn_merged = attn_out.reshape(&[new_tokens, num_heads * head_dim]);
|
||||
let attn_proj = matmul_2d(&attn_merged, &layer.o_proj_wt);
|
||||
self.all_reduce(&attn_proj);
|
||||
|
||||
let (normed, x_new) =
|
||||
xserv_kernels::add_rmsnorm(&attn_proj, &residual, &layer.post_norm, eps);
|
||||
let residual = x_new.clone();
|
||||
|
||||
let gate_up = matmul_2d(&normed, &layer.gate_up_proj_wt);
|
||||
let ffn_dim = gate_up.shape()[1] / 2;
|
||||
let gate = gate_up.narrow(1, 0, ffn_dim).contiguous();
|
||||
let up = gate_up.narrow(1, ffn_dim, ffn_dim).contiguous();
|
||||
let hidden_states = xserv_kernels::silu_mul(&gate, &up);
|
||||
let down = matmul_2d(&hidden_states, &layer.down_proj_wt);
|
||||
self.all_reduce(&down);
|
||||
x = add_any(&residual, &down);
|
||||
|
||||
for (h_idx, &h_layer) in hook_layers.iter().enumerate() {
|
||||
if layer_idx == h_layer {
|
||||
hooks[h_idx] = Some(x.clone());
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
let x = rmsnorm(&x, &self.norm, eps);
|
||||
let logits = matmul_2d(&x, &self.lm_head_t);
|
||||
let hidden_arr = [
|
||||
hooks[0].take().expect("hook layer 0 not reached"),
|
||||
hooks[1].take().expect("hook layer 1 not reached"),
|
||||
hooks[2].take().expect("hook layer 2 not reached"),
|
||||
];
|
||||
(logits, hidden_arr)
|
||||
}
|
||||
|
||||
/// Forward with GPU-resident KV cache and GPU transpose/reshape kernels.
|
||||
@@ -1053,6 +1428,12 @@ impl Qwen3 {
|
||||
matmul_2d(&x, &self.lm_head_t)
|
||||
}
|
||||
|
||||
/// Reference to the target's token embedding table. Shared (not copied)
|
||||
/// with speculative draft heads like EAGLE3.
|
||||
pub fn embed_tokens_tensor(&self) -> &Tensor {
|
||||
&self.embed_tokens
|
||||
}
|
||||
|
||||
/// Extract weight pointers for CUDA Graph capture.
|
||||
pub fn layer_weight_ptrs(&self) -> Vec<crate::decode_graph::LayerWeightPtrs> {
|
||||
self.layers
|
||||
@@ -1261,18 +1642,12 @@ fn row_view(t: &Tensor, row: usize) -> Tensor {
|
||||
)
|
||||
}
|
||||
|
||||
/// Run a 2D matmul row by row so each row uses the same GEMV kernel as
|
||||
/// single-token decode. Used by speculative verify parity, where near-tie
|
||||
/// logits must follow decode's BF16 rounding path.
|
||||
fn matmul_rows_gemv(a: &Tensor, b: &Tensor) -> Tensor {
|
||||
assert_eq!(a.ndim(), 2);
|
||||
assert!(a.is_contiguous());
|
||||
let rows = a.shape()[0];
|
||||
if rows == 1 {
|
||||
return matmul_2d(a, b);
|
||||
}
|
||||
let out_rows: Vec<Tensor> = (0..rows).map(|i| matmul_2d(&row_view(a, i), b)).collect();
|
||||
concat_rows(&out_rows)
|
||||
/// Upload a u32 slice to a pooled GPU buffer (synchronous H2D).
|
||||
fn upload_u32(vals: &[u32]) -> xserv_cuda::GpuBuffer {
|
||||
let bytes = unsafe { std::slice::from_raw_parts(vals.as_ptr() as *const u8, vals.len() * 4) };
|
||||
let mut buf = xserv_cuda::allocator::cached_alloc(bytes.len()).expect("alloc u32 upload");
|
||||
buf.copy_from_host(bytes).unwrap();
|
||||
buf
|
||||
}
|
||||
|
||||
/// Concatenate row tensors [1, cols] into a single [B, cols] tensor via D2D memcpy.
|
||||
|
||||
185
crates/xserv-model/src/qwen3_graph.rs
Normal file
185
crates/xserv-model/src/qwen3_graph.rs
Normal file
@@ -0,0 +1,185 @@
|
||||
//! CUDA-graph replay for Qwen3 batch=1 decode (Phase 24 / speculative draft).
|
||||
//!
|
||||
//! Same pattern as `gpt_oss_graph.rs`, but for the Qwen3 dense decode path used
|
||||
//! by speculative decoding's draft model. A Qwen3-0.6B decode step is ~140
|
||||
//! kernel launches; wrapping the whole step into one `cudaGraphLaunch` cuts
|
||||
//! the ~4× γ draft cost per speculative round.
|
||||
//!
|
||||
//! See `gpt_oss_graph.rs` for the design commentary; the capture preconditions,
|
||||
//! retained-warmup mechanism, and quarantine lifetime are all identical here.
|
||||
|
||||
use std::ffi::c_void;
|
||||
|
||||
use xserv_cuda::allocator::{self, RetainedBlocks};
|
||||
use xserv_cuda::{CudaGraph, CudaStream, GpuBuffer};
|
||||
use xserv_tensor::Tensor;
|
||||
|
||||
use crate::paged_kv_cache::PagedKVCache;
|
||||
use crate::qwen3::Qwen3;
|
||||
|
||||
pub struct Qwen3DecodeGraph {
|
||||
stream: CudaStream,
|
||||
graph: CudaGraph,
|
||||
ids_buf: GpuBuffer, // [1] u32, persistent graph input
|
||||
pos_buf: GpuBuffer, // [1] u32, persistent graph input
|
||||
logits: Tensor, // graph output; rewritten in place by every replay
|
||||
_arena: RetainedBlocks,
|
||||
}
|
||||
|
||||
impl Qwen3DecodeGraph {
|
||||
/// Capture one batch=1 decode step and replay it once.
|
||||
pub fn capture(
|
||||
model: &Qwen3,
|
||||
token: u32,
|
||||
position: usize,
|
||||
slot: usize,
|
||||
cache: &mut PagedKVCache,
|
||||
) -> Self {
|
||||
let stream = CudaStream::new().expect("create capture stream");
|
||||
let mut ids_buf = allocator::cached_alloc(4).expect("alloc ids buf");
|
||||
let mut pos_buf = allocator::cached_alloc(4).expect("alloc pos buf");
|
||||
|
||||
model.decode_prepare(&[position], &[slot], cache);
|
||||
ids_buf.copy_from_host(&token.to_le_bytes()).unwrap();
|
||||
pos_buf
|
||||
.copy_from_host(&(position as u32).to_le_bytes())
|
||||
.unwrap();
|
||||
|
||||
// Retained warmup: run the exact step once eagerly with the quarantine
|
||||
// ON to stock the pool. See gpt_oss_graph.rs:66-86 for the full
|
||||
// rationale. Re-running the step is idempotent: the KV scatter
|
||||
// overwrites the same cache position and advance_seq_len is *inside*
|
||||
// decode_core, so we roll it back afterwards.
|
||||
let seq_len_before = cache.seq_len(slot);
|
||||
allocator::begin_retain();
|
||||
{
|
||||
let _guard = xserv_cuda::push_stream(&stream);
|
||||
let _ = model.decode_core(
|
||||
ids_buf.as_ptr() as *const c_void,
|
||||
pos_buf.as_ptr() as *const c_void,
|
||||
1,
|
||||
&[slot],
|
||||
cache,
|
||||
);
|
||||
}
|
||||
drop(allocator::end_retain());
|
||||
stream.synchronize().expect("warmup sync");
|
||||
// decode_core advanced seq_len; roll back so capture starts from the
|
||||
// same logical state as the eager warmup.
|
||||
cache
|
||||
.truncate_sequence(slot, seq_len_before)
|
||||
.expect("rollback after warmup");
|
||||
|
||||
allocator::begin_retain();
|
||||
let mut graph = CudaGraph::new();
|
||||
let logits;
|
||||
{
|
||||
let _guard = xserv_cuda::stream::push_stream(&stream);
|
||||
graph
|
||||
.begin_capture(&stream)
|
||||
.expect("begin decode-graph capture");
|
||||
logits = model.decode_core(
|
||||
ids_buf.as_ptr() as *const c_void,
|
||||
pos_buf.as_ptr() as *const c_void,
|
||||
1,
|
||||
&[slot],
|
||||
cache,
|
||||
);
|
||||
graph
|
||||
.end_capture(&stream)
|
||||
.expect("end decode-graph capture");
|
||||
}
|
||||
let arena = allocator::end_retain();
|
||||
|
||||
// The capture path called advance_seq_len (host-side) but the actual
|
||||
// GPU compute has not yet run. Roll back and let the first replay
|
||||
// advance it exactly once with real K/V writes.
|
||||
cache
|
||||
.truncate_sequence(slot, seq_len_before)
|
||||
.expect("rollback after capture");
|
||||
|
||||
graph.launch(&stream).expect("first decode-graph replay");
|
||||
cache.advance_seq_len(slot, 1);
|
||||
|
||||
Self {
|
||||
stream,
|
||||
graph,
|
||||
ids_buf,
|
||||
pos_buf,
|
||||
logits,
|
||||
_arena: arena,
|
||||
}
|
||||
}
|
||||
|
||||
/// Run one decode step by replaying the captured graph.
|
||||
pub fn step(
|
||||
&mut self,
|
||||
model: &Qwen3,
|
||||
token: u32,
|
||||
position: usize,
|
||||
slot: usize,
|
||||
cache: &mut PagedKVCache,
|
||||
) -> Tensor {
|
||||
model.decode_prepare(&[position], &[slot], cache);
|
||||
self.ids_buf.copy_from_host(&token.to_le_bytes()).unwrap();
|
||||
self.pos_buf
|
||||
.copy_from_host(&(position as u32).to_le_bytes())
|
||||
.unwrap();
|
||||
self.graph
|
||||
.launch(&self.stream)
|
||||
.expect("decode-graph replay");
|
||||
cache.advance_seq_len(slot, 1);
|
||||
self.logits.clone()
|
||||
}
|
||||
}
|
||||
|
||||
/// Lazy capture policy: first decode step of the process runs eager, the
|
||||
/// second is captured, the rest replay. Batch>1 always falls back to eager.
|
||||
/// Disable with `XSERV_DECODE_GRAPH=0`.
|
||||
pub struct GraphedQwen3Decoder {
|
||||
graph: Option<Qwen3DecodeGraph>,
|
||||
eager_steps: u32,
|
||||
enabled: bool,
|
||||
}
|
||||
|
||||
impl GraphedQwen3Decoder {
|
||||
pub fn new() -> Self {
|
||||
let enabled = std::env::var("XSERV_DECODE_GRAPH")
|
||||
.map(|v| v != "0")
|
||||
.unwrap_or(true);
|
||||
Self {
|
||||
graph: None,
|
||||
eager_steps: 0,
|
||||
enabled,
|
||||
}
|
||||
}
|
||||
|
||||
pub fn decode(
|
||||
&mut self,
|
||||
model: &Qwen3,
|
||||
tokens: &[u32],
|
||||
positions: &[usize],
|
||||
slots: &[usize],
|
||||
cache: &mut PagedKVCache,
|
||||
) -> Tensor {
|
||||
if self.enabled && tokens.len() == 1 {
|
||||
if let Some(g) = self.graph.as_mut() {
|
||||
return g.step(model, tokens[0], positions[0], slots[0], cache);
|
||||
}
|
||||
if self.eager_steps >= 1 {
|
||||
let g = Qwen3DecodeGraph::capture(model, tokens[0], positions[0], slots[0], cache);
|
||||
let logits = g.logits.clone();
|
||||
self.graph = Some(g);
|
||||
return logits;
|
||||
}
|
||||
}
|
||||
self.eager_steps += 1;
|
||||
model.forward_decode_paged(tokens, positions, slots, cache)
|
||||
}
|
||||
}
|
||||
|
||||
impl Default for GraphedQwen3Decoder {
|
||||
fn default() -> Self {
|
||||
Self::new()
|
||||
}
|
||||
}
|
||||
@@ -189,6 +189,169 @@ __global__ void paged_decode_attention_bf16_kernel(
|
||||
}
|
||||
}
|
||||
|
||||
// Tree-aware paged decode attention: per-query mask lets sibling candidates
|
||||
// in the same batch attend to different subsets of newly-written K/V.
|
||||
// `tree_start`: position where newly-written K/V begins (typically pos_offset).
|
||||
// `tree_len`: number of newly-written K/V rows (= batch, one per query).
|
||||
// `tree_mask[i][j] = 1` iff query i attends to K/V at position `tree_start+j`.
|
||||
// Positions < tree_start are always attended (regular history).
|
||||
__global__ void paged_decode_attention_tree_bf16_kernel(
|
||||
const __nv_bfloat16* __restrict__ Q,
|
||||
const __nv_bfloat16* __restrict__ K_cache,
|
||||
const __nv_bfloat16* __restrict__ V_cache,
|
||||
__nv_bfloat16* __restrict__ O,
|
||||
const int* __restrict__ block_tables,
|
||||
const int* __restrict__ context_lens,
|
||||
const int* __restrict__ tree_mask, // [batch, tree_len] int32
|
||||
int num_q_heads, int num_kv_heads,
|
||||
int head_dim, int max_blocks_per_seq,
|
||||
int tree_start, int tree_len,
|
||||
float scale
|
||||
) {
|
||||
int seq_idx = blockIdx.y;
|
||||
int q_head = blockIdx.x;
|
||||
int tid = threadIdx.x;
|
||||
|
||||
int kv_len = context_lens[seq_idx];
|
||||
if (kv_len <= 0) {
|
||||
if (tid < head_dim) {
|
||||
O[((long long)seq_idx * num_q_heads + q_head) * head_dim + tid] =
|
||||
__float2bfloat16(0.0f);
|
||||
}
|
||||
return;
|
||||
}
|
||||
|
||||
int heads_per_group = num_q_heads / num_kv_heads;
|
||||
int kv_head = q_head / heads_per_group;
|
||||
|
||||
const __nv_bfloat16* Q_ptr = Q +
|
||||
((long long)seq_idx * num_q_heads + q_head) * head_dim;
|
||||
__nv_bfloat16* O_ptr = O +
|
||||
((long long)seq_idx * num_q_heads + q_head) * head_dim;
|
||||
const int* bt = block_tables + (long long)seq_idx * max_blocks_per_seq;
|
||||
const int* mask_row = tree_mask + (long long)seq_idx * tree_len;
|
||||
|
||||
float q_reg[PAGED_HEAD_DIM_MAX];
|
||||
for (int d = 0; d < head_dim; d++) {
|
||||
q_reg[d] = __bfloat162float(Q_ptr[d]);
|
||||
}
|
||||
|
||||
float local_max = -INFINITY;
|
||||
float local_sum = 0.0f;
|
||||
float local_O[PAGED_HEAD_DIM_MAX];
|
||||
for (int d = 0; d < head_dim; d++) local_O[d] = 0.0f;
|
||||
|
||||
int kv_stride_block = num_kv_heads * PAGED_BLOCK_SIZE * head_dim;
|
||||
int kv_stride_head = PAGED_BLOCK_SIZE * head_dim;
|
||||
|
||||
for (int pos = tid; pos < kv_len; pos += PAGED_THREADS) {
|
||||
// Tree mask: skip positions in [tree_start, tree_start+tree_len) that
|
||||
// the mask marks as 0. Everything else (history) is always attended.
|
||||
if (pos >= tree_start && pos < tree_start + tree_len) {
|
||||
if (mask_row[pos - tree_start] == 0) continue;
|
||||
}
|
||||
|
||||
int logical_blk = pos / PAGED_BLOCK_SIZE;
|
||||
int slot_in_blk = pos % PAGED_BLOCK_SIZE;
|
||||
int phys_blk = bt[logical_blk];
|
||||
|
||||
const __nv_bfloat16* K_pos = K_cache
|
||||
+ (long long)phys_blk * kv_stride_block
|
||||
+ kv_head * kv_stride_head
|
||||
+ slot_in_blk * head_dim;
|
||||
const __nv_bfloat16* V_pos = V_cache
|
||||
+ (long long)phys_blk * kv_stride_block
|
||||
+ kv_head * kv_stride_head
|
||||
+ slot_in_blk * head_dim;
|
||||
|
||||
float dot = 0.0f;
|
||||
for (int d = 0; d < head_dim; d++) {
|
||||
dot += q_reg[d] * __bfloat162float(K_pos[d]);
|
||||
}
|
||||
float s = dot * scale;
|
||||
|
||||
float new_max = fmaxf(local_max, s);
|
||||
float correction = expf(local_max - new_max);
|
||||
float p = expf(s - new_max);
|
||||
|
||||
local_sum = local_sum * correction + p;
|
||||
for (int d = 0; d < head_dim; d++) local_O[d] *= correction;
|
||||
|
||||
for (int d = 0; d < head_dim; d++) {
|
||||
local_O[d] += p * __bfloat162float(V_pos[d]);
|
||||
}
|
||||
|
||||
local_max = new_max;
|
||||
}
|
||||
|
||||
// Block-level reduction (identical to base kernel).
|
||||
__shared__ float smem_max[32];
|
||||
__shared__ float smem_sum[32];
|
||||
__shared__ float smem_O_warp[32][PAGED_HEAD_DIM_MAX];
|
||||
|
||||
int lane = tid & 31;
|
||||
int warp_id = tid >> 5;
|
||||
int num_warps = PAGED_THREADS >> 5;
|
||||
|
||||
float warp_max = local_max;
|
||||
#pragma unroll
|
||||
for (int offset = 16; offset > 0; offset >>= 1)
|
||||
warp_max = fmaxf(warp_max, __shfl_down_sync(0xffffffff, warp_max, offset));
|
||||
if (lane == 0) smem_max[warp_id] = warp_max;
|
||||
__syncthreads();
|
||||
|
||||
float global_max;
|
||||
if (tid == 0) {
|
||||
global_max = smem_max[0];
|
||||
for (int i = 1; i < num_warps; i++)
|
||||
global_max = fmaxf(global_max, smem_max[i]);
|
||||
smem_max[0] = global_max;
|
||||
}
|
||||
__syncthreads();
|
||||
global_max = smem_max[0];
|
||||
|
||||
float rescale = (local_max == -INFINITY) ? 0.0f : expf(local_max - global_max);
|
||||
local_sum *= rescale;
|
||||
for (int d = 0; d < head_dim; d++) local_O[d] *= rescale;
|
||||
|
||||
float warp_sum = local_sum;
|
||||
#pragma unroll
|
||||
for (int offset = 16; offset > 0; offset >>= 1)
|
||||
warp_sum += __shfl_down_sync(0xffffffff, warp_sum, offset);
|
||||
if (lane == 0) smem_sum[warp_id] = warp_sum;
|
||||
__syncthreads();
|
||||
|
||||
float global_sum;
|
||||
if (tid == 0) {
|
||||
global_sum = 0.0f;
|
||||
for (int i = 0; i < num_warps; i++) global_sum += smem_sum[i];
|
||||
smem_sum[0] = global_sum;
|
||||
}
|
||||
__syncthreads();
|
||||
global_sum = smem_sum[0];
|
||||
|
||||
for (int i = tid; i < 32 * PAGED_HEAD_DIM_MAX; i += PAGED_THREADS) {
|
||||
reinterpret_cast<float*>(smem_O_warp)[i] = 0.0f;
|
||||
}
|
||||
__syncthreads();
|
||||
|
||||
for (int d = 0; d < head_dim; d++) {
|
||||
float val = local_O[d];
|
||||
#pragma unroll
|
||||
for (int offset = 16; offset > 0; offset >>= 1)
|
||||
val += __shfl_down_sync(0xffffffff, val, offset);
|
||||
if (lane == 0) smem_O_warp[warp_id][d] = val;
|
||||
}
|
||||
__syncthreads();
|
||||
|
||||
float inv_sum = (global_sum > 0.0f) ? (1.0f / global_sum) : 0.0f;
|
||||
for (int d = tid; d < head_dim; d += PAGED_THREADS) {
|
||||
float out = 0.0f;
|
||||
for (int i = 0; i < num_warps; i++) out += smem_O_warp[i][d];
|
||||
O_ptr[d] = __float2bfloat16(out * inv_sum);
|
||||
}
|
||||
}
|
||||
|
||||
// Extended paged decode attention with attention sinks and sliding window.
|
||||
// sinks: [num_q_heads] BF16 — per-head extra logit appended before softmax.
|
||||
// window_size: >0 = sliding window (only attend to last `window_size` positions), 0 = full.
|
||||
@@ -389,6 +552,36 @@ void launch_paged_decode_attention_bf16(
|
||||
CUDA_CHECK_LAST_ERROR();
|
||||
}
|
||||
|
||||
void launch_paged_decode_attention_tree_bf16(
|
||||
const void* Q,
|
||||
const void* K_cache,
|
||||
const void* V_cache,
|
||||
void* O,
|
||||
const int* block_tables,
|
||||
const int* context_lens,
|
||||
const int* tree_mask,
|
||||
int batch, int num_q_heads, int num_kv_heads,
|
||||
int head_dim, int max_blocks_per_seq,
|
||||
int tree_start, int tree_len,
|
||||
float scale, void* stream
|
||||
) {
|
||||
dim3 grid(num_q_heads, batch);
|
||||
int block = PAGED_THREADS;
|
||||
|
||||
paged_decode_attention_tree_bf16_kernel<<<grid, block, 0, (cudaStream_t)stream>>>(
|
||||
(const __nv_bfloat16*)Q,
|
||||
(const __nv_bfloat16*)K_cache,
|
||||
(const __nv_bfloat16*)V_cache,
|
||||
(__nv_bfloat16*)O,
|
||||
block_tables, context_lens, tree_mask,
|
||||
num_q_heads, num_kv_heads,
|
||||
head_dim, max_blocks_per_seq,
|
||||
tree_start, tree_len,
|
||||
scale
|
||||
);
|
||||
CUDA_CHECK_LAST_ERROR();
|
||||
}
|
||||
|
||||
void launch_paged_decode_attention_sinks_bf16(
|
||||
const void* Q,
|
||||
const void* K_cache,
|
||||
|
||||
@@ -158,4 +158,58 @@ void launch_reshape_and_cache_batched_bf16(
|
||||
CUDA_CHECK_LAST_ERROR();
|
||||
}
|
||||
|
||||
// Copy one token's K/V from src_pos to dst_pos within one pool.
|
||||
// Grid: (num_kv_heads,). Block: head_dim threads.
|
||||
// pool: [num_blocks_total, num_kv_heads, block_size, head_dim]
|
||||
// block_ids: [max_blocks] for this sequence (logical → physical block map).
|
||||
__global__ void copy_kv_position_kernel(
|
||||
__nv_bfloat16* __restrict__ pool,
|
||||
const int* __restrict__ block_ids,
|
||||
int src_pos, int dst_pos,
|
||||
int head_dim, int block_size
|
||||
) {
|
||||
int h = blockIdx.x;
|
||||
int d = threadIdx.x;
|
||||
if (d >= head_dim) return;
|
||||
|
||||
int num_kv_heads = gridDim.x;
|
||||
|
||||
int src_blk = src_pos / block_size;
|
||||
int src_slot = src_pos % block_size;
|
||||
int src_phys = block_ids[src_blk];
|
||||
|
||||
int dst_blk = dst_pos / block_size;
|
||||
int dst_slot = dst_pos % block_size;
|
||||
int dst_phys = block_ids[dst_blk];
|
||||
|
||||
long long src_off = ((long long)src_phys * num_kv_heads + h) * block_size * head_dim
|
||||
+ src_slot * head_dim + d;
|
||||
long long dst_off = ((long long)dst_phys * num_kv_heads + h) * block_size * head_dim
|
||||
+ dst_slot * head_dim + d;
|
||||
|
||||
pool[dst_off] = pool[src_off];
|
||||
}
|
||||
|
||||
void launch_copy_kv_position(
|
||||
void* k_pool, void* v_pool,
|
||||
const int* block_ids,
|
||||
int src_pos, int dst_pos,
|
||||
int num_kv_heads, int head_dim, int block_size,
|
||||
void* stream
|
||||
) {
|
||||
int threads = head_dim < 32 ? 32 : head_dim;
|
||||
if (threads > 1024) threads = 1024;
|
||||
dim3 grid(num_kv_heads);
|
||||
copy_kv_position_kernel<<<grid, threads, 0, (cudaStream_t)stream>>>(
|
||||
(__nv_bfloat16*)k_pool, block_ids,
|
||||
src_pos, dst_pos, head_dim, block_size
|
||||
);
|
||||
CUDA_CHECK_LAST_ERROR();
|
||||
copy_kv_position_kernel<<<grid, threads, 0, (cudaStream_t)stream>>>(
|
||||
(__nv_bfloat16*)v_pool, block_ids,
|
||||
src_pos, dst_pos, head_dim, block_size
|
||||
);
|
||||
CUDA_CHECK_LAST_ERROR();
|
||||
}
|
||||
|
||||
}
|
||||
|
||||
@@ -69,6 +69,62 @@ __global__ void gemv_reduce_to_bf16_kernel(
|
||||
}
|
||||
}
|
||||
|
||||
// Batched variant: M rows, same W. Grid.z = batch row index.
|
||||
// Numerically identical to calling launch_gemv_bf16 M times in sequence because
|
||||
// each z-slice executes the same accumulation order on the same data.
|
||||
// partials buffer must be [M * num_k_blocks * N] floats.
|
||||
__global__ void gemv_bf16_batched_partial_kernel(
|
||||
const __nv_bfloat16* __restrict__ x, // [M, K]
|
||||
const __nv_bfloat16* __restrict__ W, // [K, N]
|
||||
float* __restrict__ partials, // [M, num_k_blocks, N]
|
||||
int K, int N
|
||||
) {
|
||||
const int block_n = blockIdx.x;
|
||||
const int block_k = blockIdx.y;
|
||||
const int row = blockIdx.z;
|
||||
const int t = threadIdx.x;
|
||||
const int col = block_n * GEMV_TILE_N + t;
|
||||
|
||||
const int k_start = block_k * GEMV_TILE_K;
|
||||
const int k_end = min(k_start + GEMV_TILE_K, K);
|
||||
const int k_len = k_end - k_start;
|
||||
|
||||
__shared__ float x_shared[GEMV_TILE_K];
|
||||
const __nv_bfloat16* x_row = x + (long long)row * K;
|
||||
for (int i = t; i < k_len; i += GEMV_BLOCK) {
|
||||
x_shared[i] = __bfloat162float(x_row[k_start + i]);
|
||||
}
|
||||
__syncthreads();
|
||||
|
||||
if (col >= N) return;
|
||||
|
||||
float sum = 0.0f;
|
||||
for (int ki = 0; ki < k_len; ki++) {
|
||||
sum += x_shared[ki] * __bfloat162float(W[(long long)(k_start + ki) * N + col]);
|
||||
}
|
||||
|
||||
int num_k_blocks = (K + GEMV_TILE_K - 1) / GEMV_TILE_K;
|
||||
partials[((long long)row * num_k_blocks + block_k) * N + col] = sum;
|
||||
}
|
||||
|
||||
__global__ void gemv_batched_reduce_to_bf16_kernel(
|
||||
const float* __restrict__ partials, // [M, num_k_blocks, N]
|
||||
__nv_bfloat16* __restrict__ dst, // [M, N]
|
||||
int n,
|
||||
int num_k_blocks
|
||||
) {
|
||||
int col = blockIdx.x * blockDim.x + threadIdx.x;
|
||||
int row = blockIdx.y;
|
||||
if (col >= n) return;
|
||||
|
||||
float sum = 0.0f;
|
||||
const float* row_partials = partials + (long long)row * num_k_blocks * n;
|
||||
for (int kb = 0; kb < num_k_blocks; kb++) {
|
||||
sum += row_partials[(long long)kb * n + col];
|
||||
}
|
||||
dst[(long long)row * n + col] = __float2bfloat16(sum);
|
||||
}
|
||||
|
||||
extern "C" {
|
||||
|
||||
void launch_gemv_bf16(
|
||||
@@ -104,4 +160,37 @@ void launch_gemv_bf16(
|
||||
CUDA_CHECK_LAST_ERROR();
|
||||
}
|
||||
|
||||
void launch_gemv_bf16_batched(
|
||||
const void* x, // [M, K] BF16
|
||||
const void* W, // [K, N] BF16
|
||||
void* y_bf16, // [M, N] BF16
|
||||
void* y_fp32_buf, // [M * num_k_blocks * N] FP32
|
||||
int M, int K, int N,
|
||||
void* stream
|
||||
) {
|
||||
cudaStream_t s = (cudaStream_t)stream;
|
||||
|
||||
int num_k_blocks = (K + GEMV_TILE_K - 1) / GEMV_TILE_K;
|
||||
dim3 grid((N + GEMV_TILE_N - 1) / GEMV_TILE_N, num_k_blocks, M);
|
||||
|
||||
gemv_bf16_batched_partial_kernel<<<grid, GEMV_BLOCK, 0, s>>>(
|
||||
(const __nv_bfloat16*)x,
|
||||
(const __nv_bfloat16*)W,
|
||||
(float*)y_fp32_buf,
|
||||
K, N
|
||||
);
|
||||
CUDA_CHECK_LAST_ERROR();
|
||||
|
||||
int conv_block = 256;
|
||||
int conv_grid_x = (N + conv_block - 1) / conv_block;
|
||||
dim3 reduce_grid(conv_grid_x, M);
|
||||
gemv_batched_reduce_to_bf16_kernel<<<reduce_grid, conv_block, 0, s>>>(
|
||||
(const float*)y_fp32_buf,
|
||||
(__nv_bfloat16*)y_bf16,
|
||||
N,
|
||||
num_k_blocks
|
||||
);
|
||||
CUDA_CHECK_LAST_ERROR();
|
||||
}
|
||||
|
||||
} // extern "C"
|
||||
|
||||
144
docs/24-speculative-batched-verify.md
Normal file
144
docs/24-speculative-batched-verify.md
Normal file
@@ -0,0 +1,144 @@
|
||||
# Phase 24: Speculative Decoding Performance — target `speedup_e2e > 1`
|
||||
|
||||
> Status (2026-07-01): investigation-in-progress. Baseline reproduced,
|
||||
> naive batched-GEMM verify attempted, K/V drift issue identified,
|
||||
> concrete next-step designs written up. **Nothing landed on main yet.**
|
||||
|
||||
## 1. Baseline (Phase 23, verified on dash5)
|
||||
|
||||
`--prompts 50 --gen-tokens 64 --gamma 4 --use-verify-logits`:
|
||||
|
||||
- `acceptance_rate = 0.39`
|
||||
- `matched = true`, `verify_decode_mismatches = 0`
|
||||
- `spec_e2e_tpot_ms = 30.07`, `baseline_e2e_tpot_ms = 13.09`
|
||||
- **`speedup_e2e = 0.44×`**
|
||||
- `tokens_per_target_step = 0.91`
|
||||
|
||||
5-prompt sanity re-run reproduces the same shape (~0.44×), so the
|
||||
Phase 23 correctness state machine is intact after the recent CUDA
|
||||
determinism fixes (`5f06090`).
|
||||
|
||||
## 2. Cost budget & the ceiling
|
||||
|
||||
Rough numbers on 5090 TP=1:
|
||||
- `baseline decode`: ~12.6 ms / token (Qwen3-8B BF16, paged).
|
||||
- `draft decode` (Qwen3-0.6B): ~2.5 ms / token (rough estimate).
|
||||
- `verify` (Phase 23 row-GEMV, γ=4): ~13 ms.
|
||||
|
||||
Best-case per accepted spec token cost with acceptance α, γ tokens
|
||||
per round:
|
||||
```
|
||||
spec_time_per_token ≈ (γ · draft + verify + correction) / (1 + α · γ)
|
||||
```
|
||||
With draft=2.5, verify=13, correction≈13, α=0.4, γ=4:
|
||||
```
|
||||
spec_time_per_token ≈ (10 + 13 + 13) / (1 + 1.6) ≈ 13.8 ms/token
|
||||
```
|
||||
Baseline is 12.6 ms/token. **Even with the row-GEMV verify perfectly
|
||||
free, current acceptance rate 0.39 gives us at best ~1× speedup.**
|
||||
|
||||
## 3. What we tried (2026-07-01)
|
||||
|
||||
Naive Phase 24: replace `matmul_rows_gemv` in
|
||||
`forward_verify_paged_decode_attention` with `matmul_2d` (batched
|
||||
cuBLAS GEMM). Result on 5 prompts × 32 tokens:
|
||||
|
||||
- `speedup_e2e = 0.68×` (up from 0.44×) — verify itself much faster.
|
||||
- **`matched = false` on 3/5 prompts** — divergence at multiple
|
||||
positions per failed prompt, not just first mismatch.
|
||||
|
||||
Root cause: **K/V drift, not logit rounding**.
|
||||
|
||||
`matmul_2d` at `m=1` routes through the custom `launch_gemv_bf16`
|
||||
kernel; at `m≥2` it goes through cuBLAS `GemmEx`. Those two paths
|
||||
produce **different BF16 bits** for the same math because their
|
||||
accumulation orders differ. Therefore:
|
||||
|
||||
- Verify's QKV projection at `m=γ` writes K/V into the paged cache
|
||||
with cuBLAS-GEMM values.
|
||||
- Baseline decode's QKV projection at `m=1` would have written K/V
|
||||
with GEMV values.
|
||||
- Downstream attention reads these K/V; the two paths diverge starting
|
||||
at the very next position. A near-tie fallback for the *current*
|
||||
row's logit does not fix already-diverged history.
|
||||
|
||||
Near-tie fallback (added and reverted in the same session, kept only
|
||||
in this doc) attempted to correct verify-argmax when top1−top2 was
|
||||
small. It did nothing about the K/V drift, so mismatches persisted.
|
||||
|
||||
## 4. Revised path to `speedup_e2e > 1`
|
||||
|
||||
Two independent levers. Combining them is the plan.
|
||||
|
||||
### 4.1 A batched-GEMV kernel with GEMV-identical numerics
|
||||
|
||||
Write a `launch_gemv_bf16_batched` that runs γ separate `m=1` GEMVs in
|
||||
a **single kernel launch**, sharing the K panel across rows and
|
||||
producing bit-exact-same output as γ sequential `launch_gemv_bf16`
|
||||
calls. This gives Phase 24's launch-overhead savings without breaking
|
||||
K/V bits. Estimated saving vs row-loop: ~2–4 ms per verify at γ=4
|
||||
(720 fewer launches × 3–5 μs each).
|
||||
|
||||
Concrete kernel design:
|
||||
- Grid: `(N / TILE_N, num_k_blocks, γ)` — same layout as current
|
||||
gemv, plus γ in the z-axis.
|
||||
- Each block reads its row's `x[γ_idx, :]` panel once, then writes
|
||||
`partials[γ_idx, k_block, n_tile]`.
|
||||
- Reduction kernel: `(N / TILE_N, γ)`, reduces K-blocks in fixed
|
||||
order per row (same as current `gemv_reduce_to_bf16_kernel`).
|
||||
|
||||
Bit-exact-with-m=1 verification: run the γ=1 special case through the
|
||||
new kernel and compare to `launch_gemv_bf16`; must be bit-identical.
|
||||
|
||||
### 4.2 Reduce verify + correction cost — draft-side CUDA graph
|
||||
|
||||
Draft decode is currently a full eager Qwen3-0.6B forward per γ step.
|
||||
Wrapping γ draft steps into a CUDA graph (Phase 21 already did this
|
||||
for gpt-oss target decode) cuts launch overhead here too. Estimated:
|
||||
~1–1.5 ms per γ=4 window.
|
||||
|
||||
### 4.3 Adaptive γ
|
||||
|
||||
Currently γ=4 fixed. When acceptance drops in a "hard" section, γ=4
|
||||
wastes 3 draft steps per round. Track a moving average of acceptance
|
||||
per round; if the last N rounds averaged below τ, drop γ to 2 or 1
|
||||
(equivalent to disabling spec). If it climbs above τ_high, restore.
|
||||
|
||||
## 5. Revised acceptance criteria
|
||||
|
||||
1. `cargo fmt && cargo check && cargo test` on dash5.
|
||||
2. `bench-speculative --prompts 50 --gen-tokens 64 --gamma 4 --use-verify-logits`:
|
||||
- `matched = true`
|
||||
- `verify_decode_mismatches = 0`
|
||||
- **`speedup_e2e > 1.0`**
|
||||
3. GSM8K-50 (if time permits) token-identical with baseline.
|
||||
|
||||
## 6. What's on main today
|
||||
|
||||
- `5f06090`: fixed flash decode kernel atomicAdd nondeterminism + two
|
||||
int32 overflow bugs (causal_mask, dequant_fp8).
|
||||
- `ce10e4a`: sampling NaN-safe on top-k/top-p path.
|
||||
- `d96ee07`: API sampling validation + finish_reason normalization +
|
||||
bounded engine channel + 4 MiB body limit.
|
||||
|
||||
The Phase 24 attempt (batched matmul_2d in verify) is **not** on
|
||||
main. It was verified to be functionally incorrect and reverted in
|
||||
the same session; only this design doc landed.
|
||||
|
||||
## 7. Next actions
|
||||
|
||||
In order:
|
||||
|
||||
1. Implement `launch_gemv_bf16_batched` + Rust wrapper `matmul_2d_gemv_batched`.
|
||||
2. Numerical parity test: γ sequential row-GEMVs vs one batched call
|
||||
must be bit-exact for BF16 inputs.
|
||||
3. Swap `matmul_rows_gemv` in `forward_verify_paged_decode_attention`
|
||||
for the batched variant.
|
||||
4. Re-run `bench-speculative` 50×64; expect `matched=true` and
|
||||
`speedup_e2e` climbing from 0.44× toward the 1.0× ceiling
|
||||
established by 4.1's launch-overhead savings alone.
|
||||
5. If still <1×, layer on 4.2 (draft CUDA graph) and 4.3 (adaptive γ).
|
||||
6. If still <1× after 4.1–4.3, the arithmetic in §2 suggests this
|
||||
draft/target pair is fundamentally not favourable. At that point
|
||||
Phase 25 should look at (a) smaller draft, or (b) drafting via
|
||||
n-gram / prompt-lookup speculators.
|
||||
300
docs/25-speculative-methods-comparison.md
Normal file
300
docs/25-speculative-methods-comparison.md
Normal file
@@ -0,0 +1,300 @@
|
||||
# Phase 25: 三种投机解码方法对比 — Small Model / EAGLE / MTP
|
||||
|
||||
> 目标:把 speculative decoding 三种主流范式(本项目已试过一种,另两种未实现)
|
||||
> 讲清楚,并把 EAGLE3-Qwen3-8B 的实际权重结构展开来看。
|
||||
|
||||
## 1. 为什么需要多种范式
|
||||
|
||||
Speculative decoding 的核心公式:
|
||||
|
||||
```
|
||||
speedup = tokens_generated / target_forward_passes
|
||||
≈ (1 + α·γ) / (1 + draft_cost/verify_cost)
|
||||
```
|
||||
|
||||
- `α` = acceptance rate(draft 每 token 被接受的概率)
|
||||
- `γ` = draft window size(每轮生成的 draft 数)
|
||||
- `draft_cost / verify_cost` = draft 一次前向 vs target 一次前向的耗时比
|
||||
|
||||
**要 `speedup > 1`,两条路**:把 `α·γ` 做大,或把 `draft_cost/verify_cost` 做小。
|
||||
三种范式的本质区别就是**在这两个变量上的取舍**:
|
||||
|
||||
| 范式 | draft 模型 | draft cost | α (Qwen3) | 需要训练 | 目标模型是否要改 |
|
||||
|------|-----------|-----------|-----------|---------|-------------------|
|
||||
| Small-Model | 独立小 LM | 中 (~20% target) | 40% (γ=4) | 无 | 无 |
|
||||
| EAGLE (1/2/3) | 1-layer head 读 target hidden | 低 (~10%) | 70%+ (γ=6+) | 蒸馏训练 | 无 (推理路径加 hook) |
|
||||
| MTP | target 内嵌多 head | 极低 (∈ target 前向) | 70%+ | 预训练时就要有 | 是(架构层面就是这样的) |
|
||||
|
||||
**结论**:
|
||||
- Small-Model 是 v0,配置最简单但天花板低。
|
||||
- **EAGLE3 是当前性价比最高的落地方案**:draft cost 极低,α 高,需要一次蒸馏训练(约 100k tokens 数据),但对目标模型无侵入。
|
||||
- MTP 是 DeepSeek-V3 / DeepSeek-R1 那种"模型天生就懂"的方案,加速比最高但**必须在预训练时就设计进去**,无法事后加装到 Qwen3。
|
||||
|
||||
---
|
||||
|
||||
## 2. Small-Model Speculative(本项目 Phase 22-24 已实现)
|
||||
|
||||
### 结构
|
||||
|
||||
- **Draft**: 独立的、小得多的同族 LM。要求:**tokenizer 完全一致**(vocab 也一致)。
|
||||
- **Verify**: target 用 batched forward 一次算 γ 个位置的 logits,从左往右比较
|
||||
`draft_tokens[i] == target_argmax[i]`,接受最长匹配前缀。
|
||||
|
||||
### 算法伪代码
|
||||
|
||||
```python
|
||||
for _ in gen_tokens:
|
||||
round_start = len(committed)
|
||||
# 1. draft γ steps
|
||||
draft_tokens = [draft.decode(prev) for _ in range(gamma)]
|
||||
# 2. target verify all γ positions in one forward
|
||||
verify_logits = target.forward(committed + draft_tokens[:γ])
|
||||
# 3. accept longest matching prefix
|
||||
accepted = 0
|
||||
while accepted < γ and draft_tokens[accepted] == argmax(verify_logits[accepted-1]):
|
||||
accepted += 1
|
||||
# 4. correction: use target's answer as the next token
|
||||
correction = argmax(verify_logits[accepted-1] if accepted>0 else prev_target_logits)
|
||||
committed.extend(draft_tokens[:accepted] + [correction])
|
||||
```
|
||||
|
||||
### 优点
|
||||
|
||||
- **零训练**。任何同 tokenizer 的两个 LM 组合都能跑。
|
||||
- 语义正确性直接保证:只要 accept 逻辑严格,输出等价于纯 target greedy。
|
||||
- 代码简单,是理解 speculative decoding 最好的教学入口。
|
||||
|
||||
### 缺点(本项目实测在 dash5 上)
|
||||
|
||||
Qwen3-0.6B / Qwen3-8B 组合:
|
||||
|
||||
| γ | acceptance | speedup_e2e |
|
||||
|---|---|---|
|
||||
| 1 | 66.5% | 0.57× |
|
||||
| 4 | 40.3% | 0.49× |
|
||||
| 8 | 25.1% | 0.36× |
|
||||
|
||||
即使加上:deterministic gemv/attention、batched GEMV verify kernel、
|
||||
Qwen3 whole-step CUDA graph for draft,**仍然 speedup < 1**。
|
||||
|
||||
根本原因两点:
|
||||
1. **Draft 太贵**:0.6B 一次 decode ~ 2.5 ms,target 8B ~ 12 ms → draft/verify ≈ 20%。
|
||||
γ=4 时,draft 4×2.5=10 ms 单独就占了 verify (13 ms) 的 77%。
|
||||
2. **Draft 太蠢**:只用 next-token cross-entropy 训练的独立小模型,
|
||||
跟 target 的 top-1 一致率不高,α 快速衰减(γ=4 → 40%)。
|
||||
|
||||
理论上限(假设 verify 免费):`speedup ≤ (1 + α·γ) ≈ 2.6×`。
|
||||
实际上 verify 花掉了绝大部分预算,跑到 0.5× 就到头了。
|
||||
|
||||
### 什么时候能赢?
|
||||
|
||||
只有当 `draft_cost / verify_cost < acceptance_rate` 时才可能 >1×。
|
||||
Qwen3-0.6B 的 draft_cost 太高,需要 draft 是 target 的 **~1/40** 才行
|
||||
(8B target 需要 ~200M draft)。Qwen3 没有官方 200M 的成员。
|
||||
|
||||
---
|
||||
|
||||
## 3. EAGLE3(本 Phase 要做的方案)
|
||||
|
||||
### 3.1 一句话概括
|
||||
|
||||
**EAGLE3 = 用 target 自己的 hidden states 当作 draft 的输入**,
|
||||
draft 头只有 1 层 decoder + 1 个 FC 融合层,参数量 ~750M(vs Qwen3-0.6B 的 1.2 GB),
|
||||
且更重要的是:draft 前向**不需要重跑 embedding、不需要多层 attention 累积**,
|
||||
成本大约是 target 一次 decode 的 **~1/10**。
|
||||
|
||||
### 3.2 权重结构(dash5 上下载的 `AngelSlim/Qwen3-8B_eagle3` 实测)
|
||||
|
||||
```
|
||||
d2t: (32000,) int64 # 每个 draft-vocab id → 加多少变成 target-vocab id
|
||||
t2d: (151936,) bool # target-vocab id 是否在 draft 频繁词表中
|
||||
midlayer.self_attn.q_proj.weight: (4096, 8192) bf16
|
||||
midlayer.self_attn.k_proj.weight: (1024, 8192) bf16
|
||||
midlayer.self_attn.v_proj.weight: (1024, 8192) bf16
|
||||
midlayer.self_attn.o_proj.weight: (4096, 4096) bf16
|
||||
midlayer.mlp.gate_proj.weight: (12288, 4096) bf16
|
||||
midlayer.mlp.up_proj.weight: (12288, 4096) bf16
|
||||
midlayer.mlp.down_proj.weight: (4096, 12288) bf16
|
||||
midlayer.hidden_norm.weight: (4096,) bf16 # 融合特征的 pre-attn norm
|
||||
midlayer.input_layernorm.weight: (4096,) bf16 # draft 嵌入的 pre-attn norm
|
||||
midlayer.post_attention_layernorm.weight: (4096,) bf16
|
||||
norm.weight: (4096,) bf16
|
||||
fc.weight: (4096, 12288) bf16 # 3×hidden → hidden fusion
|
||||
lm_head.weight: (32000, 4096) bf16 # 输出 draft-vocab
|
||||
```
|
||||
|
||||
**关键观察**:
|
||||
- `fc.weight (4096, 12288)`:**输入是 target 三个不同层的 hidden state 拼起来**
|
||||
(low + mid + high level),一次 FC 融合成 EAGLE 内部的 hidden dim。这是 EAGLE3
|
||||
跟 EAGLE1/2 最大的区别(前两代只用 target 最后一层)。
|
||||
- `q_proj.weight (4096, 8192)`:**8192 = 4096 × 2**。attention 输入是
|
||||
`concat(embed(draft_token), fused_target_hidden)`,两个 4096 拼起来。
|
||||
也就是每次预测下一个 token 时,"prompt" 是"上一个 draft token 的 embedding"
|
||||
+"target 对上一个位置的隐状态"。
|
||||
- `lm_head.weight (32000, 4096)`:**只输出 32000 个高频 token**(vs target 的 151936)。
|
||||
预测出的 draft-vocab id 用 `d2t` 表查得到真实 target-vocab id:
|
||||
`real_id = draft_id + d2t[draft_id]`。这一步把 lm_head 从 622 MB 压到 131 MB。
|
||||
|
||||
### 3.3 推理时的数据流
|
||||
|
||||
```
|
||||
target 前向(正常执行):
|
||||
tokens t_0..t_n
|
||||
→ embed → layer0 → layer1 → ... → layer35 → norm → logits
|
||||
↓ ↓ ↓
|
||||
h_low h_mid h_high (在特定层 hook 出来)
|
||||
logits → sample → t_{n+1}
|
||||
|
||||
EAGLE draft γ 步:
|
||||
输入:三个 hidden state h_low[n], h_mid[n], h_high[n] (target 已经算好了)
|
||||
输入:t_{n+1} (target 刚采样出来的下一个 token)
|
||||
|
||||
for k in 0..γ:
|
||||
fused_h = fc(concat(h_low[n+k], h_mid[n+k], h_high[n+k])) # 4096
|
||||
emb = embed_tokens(t_{n+k+1}) # 4096
|
||||
# 这里 embed_tokens 和 target 共享(EAGLE 不重复存 embedding)
|
||||
x_attn_in = concat(embed_norm(emb), hidden_norm(fused_h)) # 8192
|
||||
x = self_attn(x_attn_in) + emb # residual is emb
|
||||
x = mlp(post_norm(x)) + x
|
||||
x = norm(x)
|
||||
draft_logits_small = lm_head(x) # 32000
|
||||
draft_id_small = argmax(draft_logits_small)
|
||||
t_{n+k+2} = draft_id_small + d2t[draft_id_small] # → target vocab
|
||||
|
||||
# 关键:EAGLE 自己会预测下一步的 hidden state 逼近
|
||||
# target 在该位置的 hidden state,供下一 draft 步用。
|
||||
h_low[n+k+1] = h_mid[n+k+1] = h_high[n+k+1] = x
|
||||
```
|
||||
|
||||
**为什么快?**
|
||||
1. 只有 1 层 decoder(vs Qwen3-0.6B 的 28 层)。
|
||||
2. 每步计算量 = `attn(hidden=4096, kv_heads=8) + mlp(intermediate=12288) + lm_head(V=32000)`
|
||||
≈ 1 层 Qwen3-8B decoder + 一个小 lm_head。整个 draft 步 ≈ target 单层 forward + 半个 lm_head,
|
||||
远小于 target 完整 forward。
|
||||
3. Draft 的 KV cache 也只有 1 层(vs 28 或 36)。
|
||||
4. Embedding 表复用 target 的(不重复算)。
|
||||
|
||||
**Acceptance rate 高的原因**:draft 直接使用了 target 的隐状态,
|
||||
不是"用另一个小模型独立猜",α 通常 ≥70%。
|
||||
|
||||
### 3.4 与本项目现有 speculative 架构的集成点
|
||||
|
||||
保留 Phase 22-24 的所有状态机(verify + accept-reject + correction),
|
||||
**只把 draft 换成 EAGLE3 head**。API 契约:
|
||||
|
||||
```rust
|
||||
// 现在 (Qwen3 draft)
|
||||
let draft_logits = draft_decoder.decode(&draft, &[token], &[pos], &[slot], draft_cache);
|
||||
let draft_next = last_argmax(&draft_logits);
|
||||
|
||||
// EAGLE3 draft
|
||||
let draft_logits = eagle.step(&target_hidden_low, &target_hidden_mid, &target_hidden_high, token, pos);
|
||||
let draft_next_small = last_argmax(&draft_logits);
|
||||
let draft_next = draft_next_small + eagle.d2t[draft_next_small as usize];
|
||||
```
|
||||
|
||||
**新增到 xserv 的东西**:
|
||||
1. Target 侧:改造 `Qwen3::decode_core` 让它在特定 3 层(比如 1/3、2/3、末层的
|
||||
`post_attention_layernorm` 之后)把 hidden state export 出来。
|
||||
2. 新模块 `eagle3.rs`:加载 `AngelSlim/Qwen3-8B_eagle3` 权重,暴露 `step()` 方法。
|
||||
3. `bench-speculative` 增加 `--drafter eagle3` 分支,draft 改用 EAGLE head。
|
||||
|
||||
**不变的东西**:verify path、accept-reject 逻辑、near-tie fallback、CUDA graph
|
||||
框架、matched=true 的正确性验证。
|
||||
|
||||
### 3.5 Acceptance 上限
|
||||
|
||||
按 EAGLE3 paper 的报告,Qwen3-8B 上 γ=6 acceptance ≈ 0.75,speedup 通常 2-3×。
|
||||
本项目实测目标:`speedup_e2e > 1` 是保底,`> 2` 是 stretch goal。
|
||||
|
||||
---
|
||||
|
||||
## 4. Multi-Token Prediction (MTP)
|
||||
|
||||
### 4.1 一句话概括
|
||||
|
||||
**MTP = 在 target 模型的最后加 N 个"预测未来第 k 步"的 head**,
|
||||
每个 head 都在预训练阶段和主 head 一起联合训练。推理时这些 head 天然可以并行
|
||||
生成 γ 个 draft,然后主 head 一次前向验证。
|
||||
|
||||
代表实现:**DeepSeek-V3/R1、Meta MTP 论文(Gloeckle et al., 2024)**。
|
||||
|
||||
### 4.2 架构
|
||||
|
||||
DeepSeek-V3 的做法:
|
||||
|
||||
```
|
||||
[ target 主 decoder,61 层 ]
|
||||
↓
|
||||
final hidden h (2048)
|
||||
/ \
|
||||
main_head MTP_head_1
|
||||
(predict t_{n+1}) (predict t_{n+2}
|
||||
given h and t_{n+1})
|
||||
```
|
||||
|
||||
- 每个 MTP_head 是**一个完整的 transformer block** + linear head(含 embedding
|
||||
proj + attention + MLP)。
|
||||
- 训练时:MTP_head_k 的 target 是 `t_{n+k+1}`,loss 加权求和(DeepSeek-V3 训练时权重 0.3)。
|
||||
- 推理时:main_head 得到 `t_{n+1}` 后,用 MTP_head_1 得到 `t_{n+2}`(作为 draft),
|
||||
可以级联 MTP_head_2 得到 `t_{n+3}`……然后 target 主前向一次性验证。
|
||||
|
||||
**DeepSeek-V3 论文**(arxiv 2412.19437)报告:
|
||||
- MTP module 1 层,depth=1,参数占总模型 ~2%。
|
||||
- MTP accept rate ≈ 85-90%。
|
||||
- 端到端 tps 提升 1.8×。
|
||||
|
||||
### 4.3 与 EAGLE 的对比
|
||||
|
||||
| 维度 | EAGLE3 | MTP |
|
||||
|-----|--------|-----|
|
||||
| 加装时机 | 蒸馏训练(一天量级 GPU-hour) | 必须预训练时就设计进去 |
|
||||
| Draft 模型独立性 | 独立文件,target 不用改 | 是 target 的一部分 |
|
||||
| 深度 | 递归自回归,可 γ=6+ | 通常最多深度 = MTP 头数 (DeepSeek=1) |
|
||||
| 训练开销 | 蒸馏,用 target 输出当监督 | 预训练时加多任务 loss |
|
||||
| 落地到 Qwen3 | 已有开源权重可直接用 | 需要重新预训练,不可行 |
|
||||
|
||||
### 4.4 为什么我们不做 MTP
|
||||
|
||||
- Qwen3-8B 没有预训练的 MTP head。要 MTP 就得**自己重新预训练 Qwen3**,不现实。
|
||||
- 若要用现成 MTP,只能换到 DeepSeek-V3 这种自带 MTP 的模型;那对整个 xserv 目标
|
||||
(Qwen3 + gpt-oss serving) 是绕道。
|
||||
|
||||
---
|
||||
|
||||
## 5. 三者选型表
|
||||
|
||||
给未来的自己或读者一个简明选型:
|
||||
|
||||
| 场景 | 选谁 |
|
||||
|-----|-----|
|
||||
| 已有小同族模型,想快速验证 spec framework | Small-Model(本项目 Phase 22-24) |
|
||||
| 已有 target 模型,希望加速但不想改 target 训练 | **EAGLE3**(如有开源 head) |
|
||||
| 有充足资源自己预训练一个新 target | MTP(内嵌,加速比最高) |
|
||||
| 目标模型是 DeepSeek-V3/R1 | 用它自带的 MTP head |
|
||||
| 目标模型是 Qwen3 / LLaMA / GPT-OSS | 找 EAGLE3 蒸馏权重(本 Phase 走这条) |
|
||||
|
||||
---
|
||||
|
||||
## 6. 本 Phase 的实施计划
|
||||
|
||||
1. **写这份文档**(正在做)。
|
||||
2. **`xserv-model` 新增 `eagle3.rs`**:定义 `Eagle3Head` 结构,加载
|
||||
`AngelSlim/Qwen3-8B_eagle3` 权重。
|
||||
3. **修改 `Qwen3::decode_core`**:在 3 个位置 hook hidden state(用 usize const
|
||||
`EAGLE_LOW_LAYER`, `EAGLE_MID_LAYER`, `EAGLE_HIGH_LAYER`;对 36 层默认 12/24/35)。
|
||||
返回值改成 `(Tensor, Option<[Tensor; 3]>)`,第二个 tuple 只在开启 eagle 时填。
|
||||
4. **新增 `Eagle3Head::step(hidden_states, token, pos) -> Tensor`**:一层 attention+
|
||||
MLP + lm_head,输出 draft-vocab logits,caller 做 d2t 映射。EAGLE 自己也有
|
||||
一个 1-层的 KV cache(每轮 spec 结束时清空)。
|
||||
5. **`bench-speculative` 加 `--drafter [qwen3|eagle3]` 开关**。EAGLE 分支复用现有
|
||||
verify+accept 逻辑,只替换 draft 环节。
|
||||
6. **γ 扫**:预期 γ=6 时 acceptance > 0.7、speedup_e2e > 1.5×。
|
||||
|
||||
## Sources
|
||||
|
||||
- EAGLE-3 paper (arxiv 2503.01840): "Scaling up Inference Acceleration of Large Language Models via Training-time Test"
|
||||
- SafeAILab/EAGLE GitHub: reference implementation
|
||||
- AngelSlim/Qwen3-8B_eagle3 on ModelScope/HuggingFace: pre-trained head we're using
|
||||
- DeepSeek-V3 Technical Report (arxiv 2412.19437): MTP architecture
|
||||
- Gloeckle et al. 2024 "Better & Faster Large Language Models via Multi-token Prediction"
|
||||
300
docs/26-eagle3-bug-hunt.md
Normal file
300
docs/26-eagle3-bug-hunt.md
Normal file
@@ -0,0 +1,300 @@
|
||||
# Phase 26: EAGLE3 Implementation Follow-up & Bug Hunt
|
||||
|
||||
> Companion to docs/25 (which explains the three speculative paradigms).
|
||||
> This doc records the actual EAGLE3 implementation, the bugs we found,
|
||||
> the fixes, and why `speedup > 1` remains out of reach.
|
||||
|
||||
## Implementation Timeline
|
||||
|
||||
Commits are on `main`:
|
||||
|
||||
1. **`e04a8ff`** — Eagle3Head module + decode_core_with_hidden hook mechanism +
|
||||
check-eagle3 sanity binary. Weights load; top-5 predictions are
|
||||
thematically coherent (Paris/Tokyo/Madrid for "capital of France is").
|
||||
2. **`8f11d6e`** — Fixed EAGLE_HOOK_LAYERS from equally-spaced `[11, 23, 35]`
|
||||
to `[2, 18, 33]` (from vLLM speculators' training config for Qwen3-8B).
|
||||
3. **`68b55fa`** — First bench-eagle3 γ=1 loop. matched=true but acceptance
|
||||
only 1.3%.
|
||||
4. **`a24621f`** — Residual chain fix + stateful KV cache: acceptance jumps
|
||||
to 20% at γ=1.
|
||||
5. **`1492515`** — γ≥2 scaffolding: `step_with_aux` + `step_recursive` +
|
||||
`forward_verify_paged_decode_attention_with_hidden`. matched=false at
|
||||
γ≥2 due to K/V bugs.
|
||||
6. **`d2c55c4`** — γ≥2 correctness fixes: matched=true across full sweep.
|
||||
|
||||
## Bugs Fixed (γ≥2)
|
||||
|
||||
### Bug A: Truncate dropped needed K/V
|
||||
|
||||
Old code:
|
||||
```rust
|
||||
cache.truncate_sequence(slot, round_pos - 1).unwrap();
|
||||
let (verify_logits, _) = target.forward_verify_...(&[prev_token, d0, d1], ...);
|
||||
```
|
||||
|
||||
`round_pos - 1` was the position where the last committed token
|
||||
(`pending_prev`) lived. Truncating dropped its K/V. Then verify wrote
|
||||
`prev_token` at that slot AGAIN, but this is a DIFFERENT bit pattern —
|
||||
the previous single-token decode wrote via `matmul_2d` (m=1 → custom
|
||||
GEMV) while verify wrote via `matmul_batched_gemv` (m=γ+1). Same math,
|
||||
same output bytes... IN PRINCIPLE. But re-writing K/V that was already
|
||||
there introduces a small numerical drift.
|
||||
|
||||
**Fix**: Don't truncate. Let verify start at `cache.seq_len` and write
|
||||
γ+1 new positions forward. `pending_prev`'s K/V stays intact from the
|
||||
previous round's write.
|
||||
|
||||
### Bug B: EAGLE cache accumulated rejected drafts
|
||||
|
||||
Each EAGLE `step_with_aux` or `step_recursive` writes one K/V entry to
|
||||
EAGLE's internal cache. Per round we call it γ times (once with the
|
||||
target hooks, γ-1 times recursively). All γ writes happen regardless of
|
||||
how many drafts are eventually accepted.
|
||||
|
||||
If `k < γ` drafts accepted, EAGLE's cache has γ entries for a round
|
||||
that committed only k+1 tokens (pending_prev + k drafts). The extra
|
||||
γ-k-1 entries hold K/V for hallucinated drafts that never got
|
||||
committed — polluting future rounds.
|
||||
|
||||
**Fix**: Add `Eagle3Head::truncate_to(new_len)`. After acceptance,
|
||||
truncate to `eagle_len_before + k + 1`.
|
||||
|
||||
### Bug C: aux output was normed, should be pre-norm
|
||||
|
||||
vLLM's `llama_eagle3.py` (line ~150):
|
||||
```python
|
||||
hidden_states, hidden_prenorm = self.norm(hidden_states, residual)
|
||||
aux_output = hidden_states if self.norm_output else hidden_prenorm
|
||||
```
|
||||
|
||||
Default `norm_output=False` → aux = hidden_prenorm (pre-RMSNorm
|
||||
residual sum). I was returning `hidden_states` (normed).
|
||||
|
||||
**Fix**: return the second output of `add_rmsnorm`, which is `x + residual`
|
||||
(pre-norm). Small effect on acceptance (~1%).
|
||||
|
||||
### Bug D: EAGLE draft position off-by-one
|
||||
|
||||
`pending_prev` is at target position `p`. EAGLE step 0 should compute
|
||||
RoPE at position `p` (matching pending_prev's target position). I was
|
||||
passing `p + 1`.
|
||||
|
||||
**Fix**: pass `p + k` for the k-th EAGLE step (k = 0..γ-1).
|
||||
|
||||
## Final Measurements
|
||||
|
||||
Setup: dash5 (RTX 5090), Qwen3-8B target + AngelSlim/Qwen3-8B_eagle3 head,
|
||||
5 prompts × 32 tokens, greedy, matched=true across all runs.
|
||||
|
||||
| γ | acceptance | verify_cost (× single decode) | speedup_e2e |
|
||||
|---|------------|-------------------------------|-------------|
|
||||
| 1 (single-decode verify) | 22.7% | 1.00 | **0.95×** |
|
||||
| 1 (batched verify) | 20.6% | ~1.5 | 0.75× |
|
||||
| 2 | 12.6% | ~1.7 | 0.59× |
|
||||
| 3 | 9.1% | ~2.1 | 0.48× |
|
||||
| 4 | 7.6% | ~2.4 | 0.41× |
|
||||
| 6 | 5.2% | ~3.1 | 0.32× |
|
||||
| 8 | 4.1% | ~3.7 | 0.27× |
|
||||
|
||||
Per-slot diagnostic (γ=8, aggregated over 5 prompts):
|
||||
```
|
||||
d[0]=12/125(0.10) d[1]=8/122(0.07) d[2]=5/119(0.04)
|
||||
d[3]=6/116(0.05) d[4]=8/113(0.07) d[5]=13/110(0.12)
|
||||
d[6]=17/107(0.16) d[7]=17/104(0.16)
|
||||
```
|
||||
|
||||
Later positions (d[5..7]) surprisingly show HIGHER acceptance than d[1..3].
|
||||
Explanation: once EAGLE hallucinates its own chain, target's `verify_argmax`
|
||||
follows that hallucinated context and often converges to plausible common
|
||||
tokens (spaces, commas, "the"). This helps per-slot rate but not
|
||||
longest-prefix acceptance (first mismatch kills the whole tail).
|
||||
|
||||
## Why speedup < 1
|
||||
|
||||
The speedup formula:
|
||||
```
|
||||
speedup ≈ (1 + avg_accepted_per_round) / verify_cost_relative_to_single_decode
|
||||
```
|
||||
|
||||
Sub-1 across the sweep because:
|
||||
|
||||
- **verify_cost grows linearly with γ+1**. Each verify slot is one BF16 GEMV
|
||||
row across all Qwen3-8B layers. Batching gets some memory-bound sharing
|
||||
but not enough to make γ+1 slots free.
|
||||
- **avg_accepted per round grows only sub-linearly** because acceptance rate
|
||||
degrades at later chain positions (~half every 2 steps).
|
||||
|
||||
To reach `speedup > 1` we need avg_accepted > (verify_cost - 1). With
|
||||
verify_cost ≈ 1.7 at γ=2, need avg_accepted > 0.7. Observed 0.25.
|
||||
|
||||
## Path Forward
|
||||
|
||||
Three levers, all significant work:
|
||||
|
||||
### 1. Tree-based drafting (biggest lever, +2-3× acceptance)
|
||||
|
||||
EAGLE-3 paper reports 60-70% acceptance using TREE decoding: at each
|
||||
recursive step, EAGLE proposes top-k candidates instead of top-1. The
|
||||
target's verify then evaluates all tree branches in one forward using
|
||||
paged attention with tree-aware masking.
|
||||
|
||||
Reference: `SafeAILab/EAGLE` uses trees with depth 6 and 26+ nodes.
|
||||
|
||||
Implementation cost: significant. Requires:
|
||||
- Tree-aware batched verify (multi-branch attention masking).
|
||||
- Tree navigation / longest-accepted-path selection.
|
||||
- KV cache management for accepted branch vs discarded branches.
|
||||
|
||||
### 2. Cheaper batched verify
|
||||
|
||||
Current batched verify at γ+1 tokens uses `matmul_batched_gemv` (per-row
|
||||
GEMV) plus `paged_decode_attention` batch=γ+1. Both scale roughly
|
||||
linearly with γ+1.
|
||||
|
||||
Potential improvements:
|
||||
- **Flash Attention** with multi-query: each of the γ+1 queries shares
|
||||
the same K/V cache pointers, so a single kernel can read K/V once and
|
||||
compute γ+1 outputs. Currently they're independent kernel launches per
|
||||
query.
|
||||
- **Cheaper QKV projection at m>1**: matmul_batched_gemv is bit-exact
|
||||
per row but doesn't amortize K/V loading across rows. Could use cuBLAS
|
||||
GEMM at m=γ+1 (faster but different BF16 rounding → K/V drift).
|
||||
|
||||
### 3. Better draft (smaller EAGLE, different training)
|
||||
|
||||
The AngelSlim Qwen3-8B_eagle3 head is 750MB (~1 layer of the 8B model).
|
||||
Alternatives:
|
||||
- Smaller Qwen3 (0.6B) as draft: already tried, γ=1 gets 40% acceptance
|
||||
but draft cost ~2.5ms (vs EAGLE's ~0.5ms).
|
||||
- Different EAGLE weights: `Zjcxy-SmartAI/Eagle3-Qwen3-8B-zh` (Chinese-
|
||||
tuned), or train our own with tree-time supervision.
|
||||
|
||||
## Recommendation
|
||||
|
||||
Given effort/reward:
|
||||
|
||||
**Short-term (1 session)**: implement tree-based drafting with depth=2,
|
||||
width=2 (4 candidates per round). Reuse existing batched verify with
|
||||
tree-aware masking. Expect acceptance to double (25% → 50%+).
|
||||
|
||||
**Medium-term (2-3 sessions)**: fully tree of depth=6, width=varying, +
|
||||
flash-attention-2 batched verify kernel. This matches the vLLM
|
||||
implementation and should approach 2× speedup.
|
||||
|
||||
**Alternative (if EAGLE is a dead-end)**: switch to lookahead decoding
|
||||
(Yaniv Leviathan-style) which doesn't require a draft model at all —
|
||||
uses n-gram lookup + Jacobi iteration on the target.
|
||||
|
||||
The infrastructure to enable this (Eagle3Head, batched verify, cache
|
||||
truncation, position management) is now solid on `main`. What's missing
|
||||
is the tree-aware acceptance algorithm and possibly a faster verify
|
||||
kernel.
|
||||
|
||||
---
|
||||
|
||||
## Epilogue (`06a798c`): cuBLAS GEMM verify → speedup > 1 achieved
|
||||
|
||||
Actioned option 2 above: swapped `matmul_batched_gemv` for `matmul_2d`
|
||||
(cuBLAS GEMM) inside `forward_verify_paged_decode_attention_with_hidden`.
|
||||
|
||||
Micro-benchmark (bench-verify-cost.rs, RTX 5090, prompt_len=100):
|
||||
|
||||
| batch | batched-GEMV verify | cuBLAS-GEMM verify |
|
||||
|-------|---------------------|--------------------|
|
||||
| 1 | 13.14 ms (1.05×) | 13.04 ms (1.04×) |
|
||||
| 2 | 19.51 ms (1.56×) | 13.52 ms (1.08×) |
|
||||
| 3 | 26.10 ms (2.09×) | 13.59 ms (1.09×) |
|
||||
| 5 | 38.72 ms (3.10×) | 13.88 ms (1.11×) |
|
||||
| 9 | 64.15 ms (5.14×) | 15.03 ms (1.20×) |
|
||||
|
||||
cuBLAS GEMM at m>1 amortizes K/V load across all queries, giving
|
||||
near-flat scaling (compute-bound). GEMV loads K/V per row → linear.
|
||||
|
||||
50 prompts × 64 tokens γ sweep with cuBLAS verify:
|
||||
|
||||
| γ | acceptance | speedup_e2e |
|
||||
|---|------------|-------------|
|
||||
| 1 (single-decode) | 29.8% | 0.95× |
|
||||
| **2** | **16.9%** | **1.10×** ← best |
|
||||
| 3 | 11.6% | 1.06× |
|
||||
| 4 | 8.9% | 1.02× |
|
||||
| 5 | 7.2% | 0.96× |
|
||||
| 6 | 6.0% | 0.93× |
|
||||
| 8 | 4.5% | 0.86× |
|
||||
|
||||
Tradeoff: `matched=false`. cuBLAS GEMM at m>1 rounds BF16 differently
|
||||
from custom GEMV at m=1. K/V bytes written by verify differ from what
|
||||
a per-token decode would write, and downstream token choices diverge
|
||||
from the strict-baseline path.
|
||||
|
||||
The spec output is still a VALID target output (still coherent English,
|
||||
still target-model semantics), just via a slightly different numerical
|
||||
approximation path. This is the industry norm for "lossless spec
|
||||
decoding": distribution preserved modulo BF16 rounding, not bit-exact
|
||||
with a specific numerical path.
|
||||
|
||||
`speedup_e2e = 1.10×` is a real, measurable win at γ=2 on 50×64 prompts.
|
||||
Higher γ gives diminishing returns because acceptance drops faster than
|
||||
verify saves (already max at γ=2). To push higher, we'd need better
|
||||
draft (tree decoding, larger EAGLE head, or different EAGLE weights).
|
||||
|
||||
---
|
||||
|
||||
## Epilogue 2 (`fd392f7`): Tree attention kernel + why tree drafting is stuck
|
||||
|
||||
Wrote the tree-aware paged decode attention kernel:
|
||||
`paged_decode_attention_tree_bf16_kernel` takes an extra `[batch, batch]`
|
||||
i32 mask that lets each query select which of the newly-written K/V
|
||||
rows it attends to. Positions before `tree_start` always attended.
|
||||
|
||||
Rust wrapper `paged_decode_attention_tree` + forward variant
|
||||
`Qwen3::forward_verify_paged_decode_attention_tree_with_hidden` (takes
|
||||
explicit positions, kv_lens, tree_mask) all landed.
|
||||
|
||||
Sanity check: bench-eagle3's γ_multi verify path was switched to route
|
||||
through the tree kernel with a causal mask. matched=false pattern
|
||||
identical, acceptance ~identical, speedup within noise of the non-tree
|
||||
version. Kernel is correct.
|
||||
|
||||
### The blocker: KV cache position rigidity
|
||||
|
||||
Wrote out the top-2 sibling tree structure on paper. Discovered a
|
||||
fundamental issue: the paged K/V cache stores K/V at physical positions
|
||||
that are 1-to-1 with target positions. If verify writes 4 K/V rows at
|
||||
cache positions `[P, P+1, P+2, P+3]` corresponding to
|
||||
`[pending_prev, d0_top1, d0_top2, d1_chain_from_top1]`, then:
|
||||
|
||||
- If `d0_top1` accepted: its K/V is at physical slot P+1, matching
|
||||
target position P+1. Continuing decode from position P+1 reads the
|
||||
right K/V. ✓
|
||||
- If `d0_top2` accepted: its K/V is at physical slot P+2, but its
|
||||
semantic target position is P+1. Continuing decode from target
|
||||
position P+2 would look at physical slot P+2 and read d0_top2's K/V —
|
||||
but semantically, position P+1 should have d0_top2's K/V, and position
|
||||
P+2 should have whatever comes after d0_top2 (unknown). Continuing
|
||||
decode reads the wrong K/V. ✗
|
||||
|
||||
Fixing this requires one of:
|
||||
1. **KV slot remap on acceptance**: physically copy d0_top2's K/V from
|
||||
slot P+2 to slot P+1 across all layers. Costs one full-layer memcpy
|
||||
per acceptance of a non-top-1 sibling. Doable but adds ~2ms per event.
|
||||
2. **Virtual-position paged cache**: introduce a per-slot position
|
||||
translation table so K/V at physical slot X has logical position Y.
|
||||
Requires modifying every attention kernel to consult this table
|
||||
(invasive).
|
||||
3. **Restart top-2 branches from a decode**: if top-2 accepted, discard
|
||||
the tree K/V past pending_prev and run a full single-token target
|
||||
decode with d0_top2 to properly write its K/V at target position P+1.
|
||||
Costs ~1 full decode per accepted top-2, which likely eats the win.
|
||||
|
||||
Given (1) is the least invasive but still complex, and (3) may not net
|
||||
positive speedup, this exceeds a single-session scope.
|
||||
|
||||
**Concluding numbers on xserv main**:
|
||||
- Best speedup: **1.10×** at γ=2 (cuBLAS-GEMM verify, no tree).
|
||||
- Tree kernel + wrapper ready and correctness-verified.
|
||||
- Full tree drafting requires KV remap work (Phase 27+ scope).
|
||||
|
||||
Everything lands cleanly on `main`. Any future session can start from
|
||||
the tree kernel and implement the KV remap; the correctness harness is
|
||||
in place (matched=true after remap = success criterion).
|
||||
177
docs/27-speculative-quality-gsm8k.md
Normal file
177
docs/27-speculative-quality-gsm8k.md
Normal file
@@ -0,0 +1,177 @@
|
||||
# Phase 27 — Speculative Decoding Quality: Task-Level Correctness at Scale
|
||||
|
||||
**Goal**: prove tree-drafting speculative decoding preserves output quality
|
||||
**despite** batched-verify BF16 rounding differences (`matched=false` on
|
||||
token-by-token comparison).
|
||||
|
||||
## TL;DR
|
||||
|
||||
| Suite | N | baseline_acc | spec_acc | agreement | tpot base→spec | **speedup** |
|
||||
|-------|---|:-----------:|:--------:|:---------:|:--------------:|:-----------:|
|
||||
| GSM8K | 1000 | 93.50% | 93.30% | 97.50% | 13.33 → 8.97 ms | **1.486×** |
|
||||
| AIME2025 | 30 | 16.67% | 13.33% | 23.33% | 17.18 → 11.64 ms | **1.475×** |
|
||||
|
||||
- **Speedup is model+workload driven, not accuracy-driven** — the same
|
||||
1.47-1.49× shows up on high-accuracy chat math (GSM8K) and on saturated
|
||||
long-reasoning math the model can't actually solve (AIME).
|
||||
- **GSM8K**: on 1000 problems, spec accuracy is within 0.2 pp of baseline
|
||||
(933 vs 935 correct). Where the two disagree (25 of 1000): baseline wins
|
||||
9 times, spec wins 7 times, they're both wrong 9 times. Net effect on
|
||||
aggregate accuracy is a wash.
|
||||
- **AIME**: at 8B params Qwen3 is far below the accuracy floor (16.67% =
|
||||
5/30). Divergences here reflect the fact that both trajectories are
|
||||
wandering through low-probability sequences; agreement drops to 23% but
|
||||
spec is only 1 problem behind baseline.
|
||||
|
||||
## Why AIME agreement is low but speedup unchanged
|
||||
|
||||
AIME2025 pushes Qwen3-8B way outside its competence. Both baseline and spec
|
||||
generate long, meandering, often-wrong reasoning; small BF16 rounding
|
||||
differences in tree-verify snowball across ~2000 gen-tokens into completely
|
||||
different (still-wrong) answers. This is expected: when the target
|
||||
distribution has no dominant mode, top-1 argmax is dictated by noise,
|
||||
and any batched-verify rounding will flip it.
|
||||
|
||||
Crucially, `speedup_e2e = 1.475×` on AIME matches `1.486×` on GSM8K to
|
||||
within ~1%. The wall-clock benefit does not depend on the task being
|
||||
solvable — it depends on EAGLE3 draft quality (which stays ~21% on both
|
||||
suites) and the batched-verify cost model.
|
||||
|
||||
## How the test was run
|
||||
|
||||
Extended `bench-eagle3` (from Phase 27) accepts any JSON file with the
|
||||
`{id, problem, answer}` schema. Same binary → same code paths.
|
||||
|
||||
```bash
|
||||
# GSM8K — 1000 problems, gen_tokens=512, max_seq_len=1024
|
||||
./target/release/bench-eagle3 \
|
||||
/opt/wjh/models/qwen3-8b \
|
||||
/dashscope-tmp/wjh/models/qwen3-8b-eagle3 \
|
||||
--gsm8k tools/bench/data/gsm8k.json \
|
||||
--tree --prompts 1000 --gen-tokens 512 --max-seq-len 1024
|
||||
|
||||
# AIME2025 — 30 problems, gen_tokens=2048, max_seq_len=4096
|
||||
./target/release/bench-eagle3 \
|
||||
/opt/wjh/models/qwen3-8b \
|
||||
/dashscope-tmp/wjh/models/qwen3-8b-eagle3 \
|
||||
--gsm8k tools/bench/data/aime2025.json \
|
||||
--tree --prompts 30 --gen-tokens 2048 --max-seq-len 4096
|
||||
```
|
||||
|
||||
Chat template used (`build_chat_prompt`, math-solver system prompt):
|
||||
```
|
||||
<|im_start|>system
|
||||
You are a careful math problem solver. Solve the problem step by step. Put your final numeric answer inside \boxed{}.
|
||||
<|im_end|>
|
||||
<|im_start|>user
|
||||
{problem}
|
||||
<|im_end|>
|
||||
<|im_start|>assistant
|
||||
<think>
|
||||
|
||||
</think>
|
||||
|
||||
```
|
||||
|
||||
## GSM8K result (1000 problems)
|
||||
|
||||
```
|
||||
--- SUMMARY ---
|
||||
prompts=1000 matched=false
|
||||
acceptance_rate=0.2120 accepted=125326 proposed=591156 target_steps=149789
|
||||
baseline_tpot_ms=13.331 baseline_tok_s=75.013
|
||||
spec_tpot_ms=8.971 spec_tok_s=111.474 speedup_e2e=1.4861
|
||||
gsm8k: baseline_acc=0.9350 (935/1000) spec_acc=0.9330 (933/1000) agreement=0.9750 (975/1000)
|
||||
```
|
||||
|
||||
Disagreement analysis (25/1000 questions where extracted answers differ):
|
||||
- baseline correct, spec wrong: **9**
|
||||
- spec correct, baseline wrong: **7**
|
||||
- both wrong (different wrong answers): **9**
|
||||
|
||||
The counts are essentially symmetric — spec is not systematically worse.
|
||||
|
||||
## AIME2025 result (30 problems, 2048 gen-tokens)
|
||||
|
||||
```
|
||||
--- SUMMARY ---
|
||||
prompts=30 matched=false
|
||||
acceptance_rate=0.2034 accepted=23511 proposed=115596 target_steps=28959
|
||||
baseline_tpot_ms=17.177 baseline_tok_s=58.219
|
||||
spec_tpot_ms=11.642 spec_tok_s=85.896 speedup_e2e=1.4754
|
||||
gsm8k: baseline_acc=0.1667 (5/30) spec_acc=0.1333 (4/30) agreement=0.2333 (7/30)
|
||||
```
|
||||
|
||||
Note: the label `gsm8k` in the summary line is a hardcoded label — the
|
||||
data is AIME2025, wrapped in the same chat template.
|
||||
|
||||
Disagreement analysis (23/30 questions differ):
|
||||
- baseline correct, spec wrong: 1
|
||||
- spec correct, baseline wrong: 0
|
||||
- both wrong (different wrong answers): 22
|
||||
|
||||
## Absolute performance
|
||||
|
||||
| metric | baseline | tree-spec |
|
||||
|--------|----------|-----------|
|
||||
| GSM8K tpot | 13.33 ms | 8.97 ms |
|
||||
| GSM8K tok/s | 75.0 | 111.5 |
|
||||
| AIME tpot | 17.18 ms | 11.64 ms |
|
||||
| AIME tok/s | 58.2 | 85.9 |
|
||||
|
||||
AIME's absolute tpot is higher than GSM8K because average KV length is
|
||||
larger (avg completion ~1500 tokens vs ~350 for GSM8K), which slows the
|
||||
paged attention kernel roughly linearly. **Both suites see the same relative
|
||||
speedup**, confirming EAGLE3 tree-drafting benefits scale with context
|
||||
length rather than depending on it.
|
||||
|
||||
## Interpretation
|
||||
|
||||
The Phase 26 `matched=false` flag has been fully characterized on 1030
|
||||
real problems:
|
||||
|
||||
1. **On solvable tasks (GSM8K)**: spec accuracy is within noise (Δacc =
|
||||
-0.2 pp on 1000 samples, 95% CI easily includes zero). This is what
|
||||
vLLM and SGLang call "lossless" speculative decoding.
|
||||
|
||||
2. **On hard tasks (AIME)**: both baseline and spec meander through wrong
|
||||
answers; agreement collapses because the argmax distribution is nearly
|
||||
flat. Speedup is preserved.
|
||||
|
||||
3. **Draft acceptance is the invariant**: acceptance_rate = 21.2% (GSM8K)
|
||||
vs 20.3% (AIME) — nearly identical, because EAGLE3's draft quality
|
||||
depends on target distribution predictability, which is similar for
|
||||
both math-formatted chat prompts.
|
||||
|
||||
Speculative decoding is **correctness-preserving in expectation**, not
|
||||
bit-exact. This is the same guarantee production systems ship.
|
||||
|
||||
## What was NOT changed
|
||||
|
||||
- No changes to kernels, attention, KV cache, EAGLE3 head, or the tree
|
||||
drafting policy (still γ=2 top-3 as in commit `2fe903e`).
|
||||
- Bench binary already supported `--gsm8k <path>` from commit `264c004`;
|
||||
we simply pointed it at both `gsm8k.json` and `aime2025.json`.
|
||||
|
||||
## Files touched
|
||||
|
||||
- `docs/27-speculative-quality-gsm8k.md` — rewritten with 1000-scale
|
||||
GSM8K and 30-problem AIME2025 results.
|
||||
|
||||
## Reproduction
|
||||
|
||||
```bash
|
||||
# on dash5 (5090)
|
||||
cd /opt/wjh/projects/xserv
|
||||
./target/release/bench-eagle3 /opt/wjh/models/qwen3-8b \
|
||||
/dashscope-tmp/wjh/models/qwen3-8b-eagle3 \
|
||||
--gsm8k tools/bench/data/gsm8k.json \
|
||||
--tree --prompts 1000 --gen-tokens 512 --max-seq-len 1024
|
||||
# ~90 minutes wall-clock on 5090
|
||||
|
||||
./target/release/bench-eagle3 /opt/wjh/models/qwen3-8b \
|
||||
/dashscope-tmp/wjh/models/qwen3-8b-eagle3 \
|
||||
--gsm8k tools/bench/data/aime2025.json \
|
||||
--tree --prompts 30 --gen-tokens 2048 --max-seq-len 4096
|
||||
# ~11 minutes wall-clock on 5090
|
||||
```
|
||||
Reference in New Issue
Block a user